import os
from pathlib import Path

import numpy as np
import pandas as pd

from abc import ABC, abstractmethod
from typing import Annotated, ClassVar, Optional
import h5py

from pydantic import Field, PrivateAttr, NonNegativeFloat, PositiveFloat

from obi_one.core.block import Block
from obi_one.scientific.unions.unions_neuron_sets import NeuronSetReference
from obi_one.scientific.unions.unions_timestamps import TimestampsReference
from obi_one.scientific.circuit.circuit import Circuit
from obi_one.core.exception import OBIONE_Error



# Could be in Stimulus class rather than repeated in SomaticStimulus and SpikeStimulus
# But for now this keeps it below the other Block references in the GUI
# Eventually we can make the GUI always show the Block references at the top
_TIMESTAMPS_OFFSET_FIELD = Field(
        default=0.0, 
        title="Timestamp Offset", 
        description="The offset of the stimulus relative to each timestamp in milliseconds (ms).", 
        units="ms"
    )

class Stimulus(Block, ABC):

    timestamps: Annotated[TimestampsReference, Field(title="Timestamps", 
                                                     description="Timestamps at which the stimulus is applied.")]    

    def config(self, circuit: Circuit, population: str | None=None) -> dict:
        self.check_simulation_init()
        return self._generate_config()

    @abstractmethod
    def _generate_config(self) -> dict:
        pass


class SomaticStimulus(Stimulus, ABC):

    neuron_set: Annotated[NeuronSetReference, Field(title="Neuron Set", description="Neuron set to which the stimulus is applied.", supports_virtual=False)]

    timestamp_offset: Optional[float | list[float]] = _TIMESTAMPS_OFFSET_FIELD


    duration: NonNegativeFloat | list[NonNegativeFloat] = Field(
        default=1.0,
        title="Duration",
        description="Time duration in milliseconds for how long input is activated.",
        units="ms",
    )
    

    _represents_physical_electrode: bool = PrivateAttr(default=False) 
    """Default is False. If True, the signal will be implemented \
    using a NEURON IClamp mechanism. The IClamp produce an \
    electrode current which is not included in the calculation of \
    extracellular signals, so this option should be used to \
    represent a physical electrode. If the noise signal represents \
    synaptic input, represents_physical_electrode should be set to \
    False, in which case the signal will be implemented using a \
    MembraneCurrentSource mechanism, which is identical to IClamp, \
    but produce a membrane current, which is included in the \
    calculation of the extracellular signal."""

    def config(self, circuit: Circuit, population: str | None=None) -> dict:
        self.check_simulation_init()

        if self.neuron_set.block.population_type(circuit, population) != "biophysical":
            OBIONE_Error(
                f"Neuron Set '{self.neuron_set.block.name}' for {self.__class__.__name__}: \'{self.name}\' should be biophysical!"
            )

        return self._generate_config()


class ConstantCurrentClampSomaticStimulus(SomaticStimulus):
    """A constant current injection at a fixed absolute amplitude."""

    title: ClassVar[str] = "Constant Somatic Current Clamp (Absolute)"
    
    _module: str = "linear"
    _input_type: str = "current_clamp"

    amplitude: float | list[float] = Field(
        default=0.1, 
        description="The injected current. Given in nanoamps.",
        title="Amplitude",
        units="nA"

    )

    def _generate_config(self) -> dict:

        sonata_config = {}

        for t_ind, timestamp in enumerate(self.timestamps.block.timestamps()):
            sonata_config[self.name + "_" + str(t_ind)] = {
                "delay": timestamp + self.timestamp_offset,
                "duration": self.duration,
                "node_set": self.neuron_set.block.name,
                "module": self._module,
                "input_type": self._input_type,
                "amp_start": self.amplitude,
                "represents_physical_electrode": self._represents_physical_electrode,
            }
        return sonata_config
    
class RelativeConstantCurrentClampSomaticStimulus(SomaticStimulus):
    """A constant current injection at a percentage of each cell's threshold current."""

    title: ClassVar[str] = "Constant Somatic Current Clamp (Relative)"

    _module: str = "relative_linear"
    _input_type: str = "current_clamp"

    percentage_of_threshold_current: NonNegativeFloat | list[NonNegativeFloat] = Field(
        default=10.0,
        title="Percentage of Threshold Current",
        description="The percentage of a cell’s threshold current to inject when the stimulus \
                    activates.",
        units="%",
    )

    def _generate_config(self) -> dict:
        sonata_config = {}

        for t_ind, timestamp in enumerate(self.timestamps.block.timestamps()):
            sonata_config[self.name + "_" + str(t_ind)] = {
                "delay": timestamp + self.timestamp_offset,
                "duration": self.duration,
                "node_set": self.neuron_set.block.name,
                "module": self._module,
                "input_type": self._input_type,
                "percent_start": self.percentage_of_threshold_current,
                "represents_physical_electrode": self._represents_physical_electrode,
            }
        return sonata_config


class LinearCurrentClampSomaticStimulus(SomaticStimulus):
    """A current injection which changes linearly in absolute ampltude over time."""

    title: ClassVar[str] = "Linear Somatic Current Clamp (Absolute)"

    _module: str = "linear"
    _input_type: str = "current_clamp"

    amplitude_start: float | list[float] = Field(
        default=0.1,
        title="Start Amplitude",
        description="The amount of current initially injected when the stimulus activates. Given in nanoamps.",
        units="nA",
    )
    amplitude_end: float | list[float] = Field(
        default=0.2,
        title="End Amplitude",
        description="If given, current is interpolated such that current reaches this value when the stimulus concludes. Otherwise, current stays at \'Start Amplitude\'. Given in nanoamps.",
        units="nA",
    )

    def _generate_config(self) -> dict:
        sonata_config = {}

        for t_ind, timestamp in enumerate(self.timestamps.block.timestamps()):
            sonata_config[self.name + "_" + str(t_ind)] = {
                "delay": timestamp + self.timestamp_offset,
                "duration": self.duration,
                "node_set": self.neuron_set.block.name,
                "module": self._module,
                "input_type": self._input_type,
                "amp_start": self.amplitude_start,
                "amp_end": self.amplitude_end,
                "represents_physical_electrode": self._represents_physical_electrode,
            }
        return sonata_config


class RelativeLinearCurrentClampSomaticStimulus(SomaticStimulus):
    """A current injection which changes linearly as a percentage of each cell's threshold current over time."""

    title: ClassVar[str] = "Linear Somatic Current Clamp (Relative)"

    _module: str = "relative_linear"
    _input_type: str = "current_clamp"

    percentage_of_threshold_current_start: NonNegativeFloat | list[NonNegativeFloat] = Field(
        default=10.0,
        description="The percentage of a cell's threshold current to inject when the stimulus activates.",
        title="Percentage of Threshold Current (Start)",
        units="%",
    )
    percentage_of_threshold_current_end: NonNegativeFloat | list[NonNegativeFloat] = Field(
        default=100.0,
        description="If given, the percentage of a cell's threshold current is interpolated such that the percentage reaches this value when the stimulus concludes.",
        title="Percentage of Threshold Current (End)",
        units="%",
    )

    def _generate_config(self) -> dict:
        sonata_config = {}

        for t_ind, timestamp in enumerate(self.timestamps.block.timestamps()):
            sonata_config[self.name + "_" + str(t_ind)] = {
                "delay": timestamp + self.timestamp_offset,
                "duration": self.duration,
                "node_set": self.neuron_set.block.name,
                "module": self._module,
                "input_type": self._input_type,
                "percent_start": self.percentage_of_threshold_current_start,
                "percent_end": self.percentage_of_threshold_current_end,
                "represents_physical_electrode": self._represents_physical_electrode,
            }
        return sonata_config
    
class NormallyDistributedCurrentClampSomaticStimulus(SomaticStimulus):
    """Normally distributed current injection with a mean absolute amplitude."""

    title: ClassVar[str] = "Normally Distributed Somatic Current Clamp (Absolute)"

    _module: str = "noise"
    _input_type: str = "current_clamp"

    mean_amplitude: float | list[float] = Field(
        default=0.01, 
        description="The mean value of current to inject. Given in nanoamps (nA).",
        title="Mean Amplitude",
        units="nA"
    )
    variance: NonNegativeFloat | list[NonNegativeFloat] = Field(
        default=0.01,
        description="The variance around the mean of current to inject using a \
                    normal distribution.",
        title="Variance",
        units="nA^2"
    )

    def _generate_config(self) -> dict:
        sonata_config = {}

        for t_ind, timestamp in enumerate(self.timestamps.block.timestamps()):
            sonata_config[self.name + "_" + str(t_ind)] = {
                "delay": timestamp + self.timestamp_offset,
                "duration": self.duration,
                "node_set": self.neuron_set.block.name,
                "module": self._module,
                "input_type": self._input_type,
                "mean": self.mean_amplitude,
                "variance": self.variance,
                "represents_physical_electrode": self._represents_physical_electrode,
            }
        return sonata_config


class RelativeNormallyDistributedCurrentClampSomaticStimulus(SomaticStimulus):
    """Normally distributed current injection around a mean percentage of each cell's threshold current."""

    title: ClassVar[str] = "Normally Distributed Somatic Current Clamp (Relative)"

    _module: str = "noise"
    _input_type: str = "current_clamp"

    mean_percentage_of_threshold_current: NonNegativeFloat | list[NonNegativeFloat] = Field(
        default=0.01,
        description="The mean value of current to inject as a percentage of a cell's \
                    threshold current.",
        title="Percentage of Threshold Current (Mean)",
        units="%",
    )
    variance: NonNegativeFloat | list[NonNegativeFloat] = Field(
        default=0.01,
        description="The variance around the mean of current to inject using a \
                    normal distribution.",
        title="Variance",
        units="nA^2",
    )

    def _generate_config(self) -> dict:
        sonata_config = {}

        for t_ind, timestamp in enumerate(self.timestamps.block.timestamps()):
            sonata_config[self.name + "_" + str(t_ind)] = {
                "delay": timestamp + self.timestamp_offset,
                "duration": self.duration,
                "node_set": self.neuron_set.block.name,
                "module": self._module,
                "input_type": self._input_type,
                "mean_percent": self.mean_percentage_of_threshold_current,
                "variance": self.variance,
                "represents_physical_electrode": self._represents_physical_electrode,
            }
        return sonata_config


class MultiPulseCurrentClampSomaticStimulus(SomaticStimulus):
    """A series of current pulses injected at a fixed frequency, with each pulse having a fixed absolute amplitude and temporal width."""

    title: ClassVar[str] = "Multi Pulse Somatic Current Clamp (Absolute)"

    _module: str = "pulse"
    _input_type: str = "current_clamp"

    amplitude: float | list[float] = Field(
        default=0.1,
        description="The amount of current initially injected when each pulse activates. Given in nanoamps (nA).",
        title="Amplitude",
        units="nA",
    )
    width: float | list[float] = Field(
        default=1.0, 
        description="The length of time each pulse lasts. Given in milliseconds (ms).",
        title="Pulse Width",
        units="ms"
    )
    frequency: float | list[float] = Field(
        default=1.0, 
        description="The frequency of pulse trains. Given in Hertz (Hz).",
        title="Pulse Frequency",
        units="Hz"
    )

    def _generate_config(self) -> dict:
        sonata_config = {}

        for t_ind, timestamp in enumerate(self.timestamps.block.timestamps()):
            sonata_config[self.name + "_" + str(t_ind)] = {
                "delay": timestamp + self.timestamp_offset,
                "duration": self.duration,
                "node_set": self.neuron_set.block.name,
                "module": self._module,
                "input_type": self._input_type,
                "amp_start": self.amplitude,
                "width": self.width,
                "frequency": self.frequency,
                "represents_physical_electrode": self._represents_physical_electrode,
            }
        return sonata_config


class SinusoidalCurrentClampSomaticStimulus(SomaticStimulus):
    """A sinusoidal current injection with a fixed frequency and maximum absolute amplitude."""

    title: ClassVar[str] = "Sinusoidal Somatic Current Clamp (Absolute)"

    _module: str = "sinusoidal"
    _input_type: str = "current_clamp"

    maximum_amplitude: float | list[float] = Field(
        default=0.1, 
        description="The maximum (and starting) amplitude of the sinusoid. Given in nanoamps (nA).",
        title="Maximum Amplitude",
        units="nA"
    )
    frequency: float | list[float] = Field(
        default=1.0, 
        description="The frequency of the waveform. Given in Hertz (Hz).",
        title="Frequency",
        units="Hz"
    )
    dt: float | list[float] = Field(
        default=0.025, 
        description="Timestep of generated signal in milliseconds (ms).",
        title="Timestep",
        units="ms"
    )

    def _generate_config(self) -> dict:
        sonata_config = {}

        for t_ind, timestamp in enumerate(self.timestamps.block.timestamps()):
            sonata_config[self.name + "_" + str(t_ind)] = {
                "delay": timestamp + self.timestamp_offset,
                "duration": self.duration,
                "node_set": self.neuron_set.block.name,
                "module": self._module,
                "input_type": self._input_type,
                "amp_start": self.peak_amplitude,
                "frequency": self.frequency,
                "dt": self.dt,
                "represents_physical_electrode": self._represents_physical_electrode,
            }
        return sonata_config


class SubthresholdCurrentClampSomaticStimulus(SomaticStimulus):
    """A subthreshold current injection at a percentage below each cell's threshold current."""

    title: ClassVar[str] = "Subthreshold Somatic Current Clamp (Relative)"

    _module: str = "subthreshold"
    _input_type: str = "current_clamp"

    percentage_below_threshold: float | list[float] = Field(
        default=0.1,
        description="A percentage adjusted from 100 of a cell's threshold current. \
                        E.g. 20 will apply 80\% of the threshold current. Using a negative \
                            value will give more than 100. E.g. -20 will inject 120\% of the \
                                threshold current.",
        title="Percentage Below Threshold",
        units="%",
    )

    def _generate_config(self) -> dict:
        sonata_config = {}

        for t_ind, timestamp in enumerate(self.timestamps.block.timestamps()):
            sonata_config[self.name + "_" + str(t_ind)] = {
                "delay": timestamp + self.timestamp_offset,
                "duration": self.duration,
                "node_set": self.neuron_set.block.name,
                "module": self._module,
                "input_type": self._input_type,
                "percent_less": self.percentage_below_threshold,
                "represents_physical_electrode": self._represents_physical_electrode,
            }
        return sonata_config


class HyperpolarizingCurrentClampSomaticStimulus(SomaticStimulus):
    """A hyperpolarizing current injection which brings a cell to base membrance voltage. \
        The holding current is pre-defined for each cell.
    """

    title: ClassVar[str] = "Hyperpolarizing Somatic Current Clamp"

    _module: str = "hyperpolarizing"
    _input_type: str = "current_clamp"

    def _generate_config(self) -> dict:
        sonata_config = {}

        for t_ind, timestamp in enumerate(self.timestamps.block.timestamps()):
            sonata_config[self.name + "_" + str(t_ind)] = {
                "delay": timestamp + self.timestamp_offset,
                "duration": self.duration,
                "node_set": self.neuron_set.block.name,
                "module": self._module,
                "input_type": self._input_type,
                "represents_physical_electrode": self._represents_physical_electrode,
            }
        return sonata_config


class SpikeStimulus(Stimulus):
    _module: str = "synapse_replay"
    _input_type: str = "spikes"
    _spike_file: Path | None = None
    _simulation_length: float | None = None
    source_neuron_set: Annotated[NeuronSetReference, Field(title="Neuron Set (Source)", supports_virtual=True)]
    targeted_neuron_set: Annotated[NeuronSetReference, Field(title="Neuron Set (Target)", supports_virtual=False)]

    timestamp_offset: Optional[float | list[float]] = _TIMESTAMPS_OFFSET_FIELD

    def config(self, circuit: Circuit, population: str | None=None) -> dict:
        self.check_simulation_init()

        if self.targeted_neuron_set.block.population_type(circuit, population) != "biophysical":
            raise OBIONE_Error(
                f"Target Neuron Set '{self.targeted_neuron_set.block.name}' for {self.__class__.__name__}: \'{self.name}\' should be biophysical!"
            )

        return self._generate_config()

    def _generate_config(self) -> dict:
        assert self._spike_file is not None
        assert self._simulation_length is not None, "Simulation length must be set before generating SONATA config component for SpikeStimulus."
        # assert self.source_neuron_set.block.node_population is not None, "Must specify node population name for the neuron set!"
        sonata_config = {}
        sonata_config[self.name] = {
                "delay": 0.0, # If it is present, then the simulation filters out those times that are before the delay
                "duration": self._simulation_length,
                "node_set": self.targeted_neuron_set.block.name,
                "module": self._module,
                "input_type": self._input_type,
                "spike_file": str(self._spike_file) # os.path.relpath # 
            }
        
        return sonata_config

    def generate_spikes(self, circuit, spike_file_path, simulation_length, source_node_population=None):
        raise NotImplementedError("Subclasses should implement this method.")

    @staticmethod
    def write_spike_file(gid_spike_map, spike_file, source_node_population):
        """
        Writes SONATA output spike trains to file.
        
        Spike file format specs: https://github.com/AllenInstitute/sonata/blob/master/docs/SONATA_DEVELOPER_GUIDE.md#spike-file
        """
        # IMPORTANT: Convert SONATA node IDs (0-based) to NEURON cell IDs (1-based)!!
        # (See https://sonata-extension.readthedocs.io/en/latest/blueconfig-projection-example.html#dat-spike-files)
        gid_spike_map = {k + 1: v for k, v in gid_spike_map.items()}
    
        out_path = os.path.split(spike_file)[0]
        if not os.path.exists(out_path):
            os.makedirs(out_path)

        time_list = []
        gid_list = []
        for gid, spike_times in gid_spike_map.items():
            if spike_times is not None:
                for t in spike_times:
                    time_list.append(t)
                    gid_list.append(gid)
        spike_df = pd.DataFrame(np.array([time_list, gid_list]).T, columns=['t', 'gid'])
        spike_df = spike_df.astype({'t': float, 'gid': int})
        spike_df.sort_values(by=['t', 'gid'], inplace=True)  # Sort by time
        with h5py.File(spike_file, 'w') as f:
            pop = f.create_group(f"/spikes/{source_node_population}")
            ts = pop.create_dataset("timestamps", data=spike_df['t'].values, dtype=np.float64)
            nodes = pop.create_dataset("node_ids", data=spike_df['gid'].values, dtype=np.uint64)
            ts.attrs['units'] = 'ms'

class PoissonSpikeStimulus(SpikeStimulus):
    """Spike times drawn from a Poisson process with a given frequency \
        sent from all neurons in the source neuron set to efferently connected \
            neurons in the target neuron set."""

    title: ClassVar[str] = "Poisson Spikes (Efferent)"

    _module: str = "synapse_replay"
    _input_type: str = "spikes"
    duration: NonNegativeFloat | list[NonNegativeFloat] = Field(default=1000.0, 
                            title="Duration", 
                            description="Time duration in milliseconds for how long input is activated.",
                            units="ms"
                        )
    frequency: PositiveFloat | list[PositiveFloat] = Field(default=1.0, 
                                           title="Frequency", 
                                           description="Mean frequency (Hz) of the Poisson input.",
                                           units="Hz")
    random_seed: int | list[int] = Field(
        default=0,
        title="Random Seed",
        description="Seed for the random number generator to ensure reproducibility of the spike generation.",
    )
    
    def generate_spikes(self, circuit, spike_file_path, simulation_length, source_node_population=None):
        self._simulation_length = simulation_length
        rng = np.random.default_rng(self.random_seed)
        gids = self.source_neuron_set.block.get_neuron_ids(circuit, source_node_population)
        source_node_population = self.source_neuron_set.block._population(source_node_population)
        gid_spike_map = {}
        timestamps = self.timestamps.block.timestamps()
        for timestamp_idx, timestamp_t in enumerate(timestamps):
            start_time = timestamp_t + self.timestamp_offset
            end_time = start_time + self.duration
            if timestamp_idx < len(timestamps) - 1:
                # Check that interval not overlapping with next stimulus onset
                assert end_time < timestamps[timestamp_idx + 1], "Stimulus time intervals overlap!"
            for gid in gids:
                spikes = []
                t = start_time
                while t < end_time:
                    # Draw next spike time from exponential distribution
                    interval = rng.exponential(1.0 / self.frequency) * 1000  # convert s → ms
                    t += interval
                    if t < end_time:
                        spikes.append(t)
                if gid in gid_spike_map:
                    gid_spike_map[gid] = gid_spike_map[gid] + spikes
                else:
                    gid_spike_map[gid] = spikes
        self._spike_file = f"{self.name}_spikes.h5"
        self.write_spike_file(gid_spike_map, spike_file_path / self._spike_file, source_node_population)


class FullySynchronousSpikeStimulus(SpikeStimulus):
    """Spikes sent at the same time from all neurons in the source neuron set \
        to efferently connected neurons in the target neuron set."""

    title: ClassVar[str] = "Fully Synchronous Spikes (Efferent)"

    _module: str = "synapse_replay"
    _input_type: str = "spikes"

    def generate_spikes(self, circuit, spike_file_path, simulation_length, source_node_population=None):
        self._simulation_length = simulation_length
        gids = self.source_neuron_set.block.get_neuron_ids(circuit, source_node_population)
        source_node_population = self.source_neuron_set.block._population(source_node_population)
        gid_spike_map = {}
        timestamps = self.timestamps.block.timestamps()
        for t_idx, start_time in enumerate(timestamps):
            spike = [start_time + self.timestamp_offset]
            for gid in gids:
                if gid in gid_spike_map:
                    gid_spike_map[gid] = gid_spike_map[gid] + spike
                else:
                    gid_spike_map[gid] = spike
        self._spike_file = f"{self.name}_spikes.h5"
        self.write_spike_file(gid_spike_map, spike_file_path / self._spike_file, source_node_population)




from abc import ABC, abstractmethod
from typing import Annotated, ClassVar

from pydantic import Field, NonNegativeFloat, NonNegativeInt

from obi_one.core.block import Block


class Timestamps(Block, ABC):
    start_time: Annotated[
        NonNegativeFloat | list[NonNegativeFloat], Field(default=0.0, description="Sart time of the timestamps in milliseconds (ms).", units="ms")
    ]

    def timestamps(self):
        self.check_simulation_init()
        return self._resolve_timestamps()

    @abstractmethod
    def _resolve_timestamps(self):
        pass


class SingleTimestamp(Timestamps):
    """A single timestamp at a specified time."""

    title: ClassVar[str] = "Single Timestamp"

    def _resolve_timestamps(self) -> list[float]:
        return [self.start_time]


class RegularTimestamps(Timestamps):
    """A series of timestamps at regular intervals."""

    title: ClassVar[str] = "Regular Timestamps"

    interval: Annotated[
        NonNegativeFloat | list[NonNegativeFloat], Field(default=10.0, description="Interval between timestamps in milliseconds (ms).", units="ms")
    ]

    number_of_repetitions: Annotated[
        NonNegativeInt | list[NonNegativeInt], Field(default=10, description="Number of timestamps to generate.")
    ]

    def _resolve_timestamps(self) -> list[float]:
        return [self.start_time + i * self.interval for i in range(self.number_of_repetitions)]



import re
from typing import Annotated, get_type_hints
import tempfile

import entitysdk.client
import entitysdk.common
from entitysdk.exception import EntitySDKError
from fastapi import APIRouter, Depends, HTTPException
from fastapi.responses import JSONResponse

from app.config import settings
from app.dependencies.entitysdk import get_client
from app.logger import L
from obi_one import *
from obi_one.core.form import Form
from obi_one.core.scan import GridScan


def check_implementations_of_single_coordinate_class(
    model: type[Form], processing_method: str, data_postprocessing_method: str
) -> str | type | None:
    """Return the class of the return type of a processing_method of the single coordinate class.

    Returns None if return type not specified
    Returns message strings if the processing_method
    or data_postprocessing_method not implemented.
    """
    return_class = None

    # Check that the single_coord_class_name is set
    if not model.single_coord_class_name:
        return f"single_coord_class_name is not set in the form: {model.__name__}"
    single_coordinate_cls = globals().get(model.single_coord_class_name)
    if single_coordinate_cls is None:
        return f"Class {model.single_coord_class_name} not found in globals"

    # Check that the method is a method of the single coordinate class
    if not (
        hasattr(single_coordinate_cls, processing_method)
        and callable(getattr(single_coordinate_cls, processing_method))
    ):
        return f"{processing_method} is not a method of {single_coordinate_cls.__name__}"
    if not data_postprocessing_method:
        return None

    # Check that the data_postprocessing_method is a method of the single coordinate class
    if not (
        hasattr(single_coordinate_cls, data_postprocessing_method)
        and callable(getattr(single_coordinate_cls, data_postprocessing_method))
    ):
        return f"{data_postprocessing_method} is not a method of {single_coordinate_cls.__name__}"
    return_class = get_type_hints(getattr(single_coordinate_cls, data_postprocessing_method)).get(
        "return"
    )

    return return_class


def create_endpoint_for_form(
    model: type[Form], router: APIRouter, processing_method: str, data_postprocessing_method: str
) -> None:
    """Create a FastAPI endpoint for generating grid scans based on an OBI Form model."""
    # model_name: model in lowercase with underscores between words and "Forms" removed (i.e.
    # 'morphology_metrics_example')
    model_base_name = model.__name__.removesuffix("Form")
    model_name = "-".join([word.lower() for word in re.findall(r"[A-Z][^A-Z]*", model_base_name)])

    # Check which of single coordinate class, method, data_handling_method and return type
    # are implemented
    return_class = check_implementations_of_single_coordinate_class(
        model, processing_method, data_postprocessing_method
    )

    if not isinstance(return_class, str):
        # return_type = None if return_class is None else dict[str, return_class]

        # Create endpoint name
        endpoint_name_with_slash = "/" + model_name + "-" + processing_method + "-grid"
        if data_postprocessing_method:
            endpoint_name_with_slash = endpoint_name_with_slash + "-" + data_postprocessing_method

        @router.post(endpoint_name_with_slash, summary=model.name, description=model.description)
        def endpoint(
            db_client: Annotated[entitysdk.client.Client, Depends(get_client)],
            form: model,
        ) -> str:
            L.info("generate_grid_scan")
            L.info(db_client)

            campaign = None
            try:
                with tempfile.TemporaryDirectory() as tdir:
                    grid_scan = GridScan(
                        form=form,
                        # output_root=settings.OUTPUT_DIR / "fastapi_test" / model_name / "grid_scan",
                        output_root=tdir,
                        coordinate_directory_option="ZERO_INDEX",
                    )
                    campaign = grid_scan.execute(
                        processing_method=processing_method,
                        data_postprocessing_method=data_postprocessing_method,
                        db_client=db_client,
                    )

            except Exception as e:
                error_msg = str(e)

                if len(e.args) == 1:
                    error_msg = str(e.args[0])
                elif len(e.args) > 1:
                    error_msg = str(e.args)

                L.error(error_msg)

                raise HTTPException(status_code=500, detail=error_msg) from e

            else:
                L.info("Grid scan generated successfully")
                if campaign is not None:
                    return str(campaign.id)

                L.info("No campaign generated")
                return ""


def activate_generated_endpoints(router: APIRouter) -> APIRouter:
    # 1. Create endpoints for each OBI Form subclass.
    for form, processing_method, data_postprocessing_method in [
        (SimulationsForm, "generate", ""),
        (SimulationsForm, "generate", "save"),
        (MorphologyMetricsForm, "run", ""),
    ]:
        # Create endpoint
        create_endpoint_for_form(
            form,
            router,
            processing_method=processing_method,
            data_postprocessing_method=data_postprocessing_method,
        )

    return router



from pydantic import PrivateAttr
from typing import Optional

from obi_one.core.base import OBIBaseModel
from obi_one.core.param import MultiValueScanParam

from typing import ClassVar


class Block(OBIBaseModel):
    """Defines a component of a Form.

    Parameters can be of type | list[type]
    when a list is used it is used as a dimension in a multi-dimensional parameter scan.
    Tuples should be used when list-like parameter is needed.
    """

    title: ClassVar[Optional[str]] = None  # Optional: subclasses can override

    @classmethod
    def __init_subclass__(cls, **kwargs) -> None:
        super().__init_subclass__(**kwargs)

        # Use the subclass-provided title, or fall back to the class name
        cls.model_config = {
            "title": cls.title or cls.__name__
        }

    _multiple_value_parameters: list[MultiValueScanParam] = PrivateAttr(default=[])

    _simulation_level_name: Optional[str] = PrivateAttr(default=None)

    _ref = None

    def check_simulation_init(self):
        assert self._simulation_level_name is not None, (
            f"'{self.__class__.__name__}' initialization within a simulation required!"
        )

    @property
    def name(self):
        self.check_simulation_init()
        return self._simulation_level_name

    def has_name(self) -> bool:
        return self._simulation_level_name is not None

    def set_simulation_level_name(self, value: str):
        if not isinstance(value, str) or not value:
            raise ValueError("Simulation level name must be a non-empty string.")
        self._simulation_level_name = value

    @property
    def ref(self):
        if self._ref is None:
            raise ValueError("Block reference has not been set.")
        return self._ref
    
    def set_ref(self, value):
        self._ref = value


    def multiple_value_parameters(
        self, category_name: str, block_key: str = ""
    ) -> list[MultiValueScanParam]:
        """Return a list of MultiValueScanParam objects for the block."""
        self._multiple_value_parameters = []

        for key, value in self.__dict__.items():
            if isinstance(value, list):  # and len(value) > 1:
                multi_values = value
                if block_key:
                    self._multiple_value_parameters.append(
                        MultiValueScanParam(
                            location_list=[category_name, block_key, key], values=multi_values
                        )
                    )
                else:
                    self._multiple_value_parameters.append(
                        MultiValueScanParam(location_list=[category_name, key], values=multi_values)
                    )

        return self._multiple_value_parameters

    def enforce_no_lists(self) -> None:
        """Raise a TypeError if any attribute is a list."""
        for key, value in self.__dict__.items():
            if isinstance(value, list):
                msg = f"Attribute '{key}' must not be a list."
                raise TypeError(msg)



import abc
from typing import ClassVar

from pydantic import BaseModel


class Activity(BaseModel, abc.ABC):
    """Base class for validation objects.

    This class is used to define the structure of validation objects.
    It can be extended to create specific validation types.
    """

    name: ClassVar[str]
    description: ClassVar[str]

    def run(self) -> None:
        """Validate the provided data against the validation rules."""
        raise NotImplementedError("Subclasses must implement this method.")

    def save(self) -> None:
        """Make a call to entitysdk to save the result of the validation."""
        raise NotImplementedError("Subclasses must implement this method.")


from pathlib import Path
from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, ValidatorFunctionWrapHandler, model_validator


def get_subclasses_recursive[T](cls: type[T]) -> list[type[T]]:
    """Returns all the subclasses of a given class."""
    subclasses = []
    for subclass in cls.__subclasses__():
        subclasses.append(subclass)
        subclasses.extend(get_subclasses_recursive(subclass))
    return subclasses


def get_subclass_recursive[T](cls: type[T], name: str) -> type[T]:
    return next(c for c in get_subclasses_recursive(cls=cls) if c.__qualname__ == name)


class OBIBaseModel(BaseModel):
    """Sets `type` fields for model_dump which are then used for desserialization.

    Sets encoder for EntitySDK Entities
    """

    type: str = ""

    model_config = ConfigDict(json_encoders={Path: str})

    @model_validator(mode="before")
    @classmethod
    def set_type(cls, data: Any) -> dict[str, Any]:
        """Automatically sets `type` when instantiated in Python if a dictionary."""
        if isinstance(data, dict) and "type" not in data:
            data["type"] = cls.__qualname__
        return data

    def __init_subclass__(cls, **kwargs) -> None:
        """Dynamically set the `type` field to the class name."""
        super().__init_subclass__(**kwargs)
        cls.__annotations__["type"] = Literal[cls.__qualname__]

    def __str__(self) -> str:
        """Return a string representation of the OBIBaseModel object."""
        return self.__repr__()


    @model_validator(mode="wrap")
    @classmethod
    def retrieve_type_on_deserialization(cls, 
                                        value: Any, 
                                        handler: ValidatorFunctionWrapHandler
                ) -> "OBIBaseModel":

        if isinstance(value, dict):

            # `sub_cls(**modified_value)` will trigger a recursion, and thus we need to
            # remove `type` from the dictionary before passing it to the subclass constructor
            
            modified_value = value.copy()
            sub_cls_name = modified_value.pop("type", None)
            
            if sub_cls_name is not None:

                sub_cls = get_subclass_recursive(
                    cls=OBIBaseModel,
                    name=sub_cls_name,  # , allow_same_class=True
                )

                return sub_cls(**modified_value)

            return handler(value)

        return handler(value)



import obi_one as obi


def test_simulation_campaign():
    print("Testing Simulation Campaign... To be filled in later.")
    # circuit_path_prefix = "/Users/james/Documents/obi/additional_data/"

    # circuit = obi.Circuit(name="ToyCircuit-S1-6k", path=circuit_path_prefix + "ToyCircuit-S1-6k/circuit_config.json")
    # print(f"Circuit '{circuit}' with {circuit.sonata_circuit.nodes.size} neurons and {circuit.sonata_circuit.edges.size} synapses")

    # circuit2 = obi.Circuit(name="nbS1-HEX0-beta", path=circuit_path_prefix + "ToyCircuit-S1-6k/circuit_config.json")
    # print(f"Circuit '{circuit2}' with {circuit2.sonata_circuit.nodes.size} neurons and {circuit2.sonata_circuit.edges.size} synapses")

    # Simulation init
    # sim_neuron_set = obi.PredefinedNeuronSet(node_set="Layer1")
    # sim_duration = 3000.0  # ms
    # simulations_initialize = obi.SimulationsForm.Initialize(
    #     circuit=[circuit, circuit2],
    #     node_set=sim_neuron_set,
    #     simulation_length=sim_duration,
    # )

    # # Stimuli
    # stim_neuron_set = obi.PredefinedNeuronSet(node_set="Layer1", sample_percentage=[10, 20])
    # stim_times = obi.RegularTimestamps(
    #     start_time=0.0, number_of_repetitions=3, interval=1000.0
    # )  # in ms!!
    # current_stimulus = obi.ConstantCurrentClampSomaticStimulus(
    #     timestamps=stim_times, duration=5.0, neuron_set=stim_neuron_set, amplitude=[0.2, 0.5]
    # )
    # sync_current_stimulus = obi.ConstantCurrentClampSomaticStimulus(
    #     timestamps=stim_times, duration=100.0, neuron_set=stim_neuron_set, amplitude=0.1
    # )

    # # Recordings
    # rec_neuron_set = obi.PredefinedNeuronSet(node_set="Layer1", sample_percentage=100)
    # v_recording = obi.SomaVoltageRecording(
    #     start_time=0.0, end_time=sim_duration, neuron_set=rec_neuron_set
    # )

    # """
    # Fill form with Blocks
    # """
    # simulations_form = obi.SimulationsForm(
    #                                     initialize=simulations_initialize,
    #                                     intracellular_location_sets={},
    #                                     extracellular_location_sets={},
    #                                     neuron_sets={"L1All": sim_neuron_set, "L1Stim": stim_neuron_set, "L1Rec": rec_neuron_set},
    #                                     synapse_sets={},
    #                                     timestamps={"StimTimes": stim_times},
    #                                     stimuli={"CurrentStimulus": current_stimulus, "SyncCurrentStimulus": sync_current_stimulus},
    #                                     recordings={"SomaVoltRec": v_recording},
    #                                     )

    # simulations_form.model_dump(mode="json")

    # grid_scan_output_path = Path('../obi-output-tests/circuit_simulations/grid_scan')
    # if grid_scan_output_path.exists():
    #     grid_scan_output_path.rmdir()

    # grid_scan = obi.GridScan(form=simulations_form, output_root=grid_scan_output_path)
    # grid_scan.multiple_value_parameters(display=True)
    # grid_scan.coordinate_parameters(display=True)
    # grid_scan.coordinate_instances(display=True)
    # grid_scan.execute(processing_method='generate')

    # # Deserialization
    # grid_scan_ds = obi.deserialize_obi_object_from_json_file(grid_scan_output_path / "run_scan_config.json")



from obi_one.core.base import OBIBaseModel
from obi_one.core.block import Block
from obi_one.core.block_reference import BlockReference
from obi_one.core.form import Form
from obi_one.core.info import Info
from obi_one.core.path import NamedPath
from obi_one.core.scan import CoupledScan, GridScan
from obi_one.core.serialization import (
    deserialize_obi_object_from_json_data,
    deserialize_obi_object_from_json_file,
)
from obi_one.core.single import SingleCoordinateMixin
from obi_one.core.tuple import NamedTuple
from obi_one.database.db_manager import db
from obi_one.core.activity import Activity
from obi_one.core.validation import Validation
from obi_one.core.exception import OBIONE_Error

__all__ = [
    "Activity",
    "AllNeurons",
    "AfferentSynapsesBlock",
    "BasicConnectivityPlot",
    "BasicConnectivityPlots",
    "Block",
    "BlockReference",
    "rCA1CA3Inputs",
    "Circuit",
    "CircuitExtraction",
    "CircuitExtractions",
    "CircuitFromID",
    "ClusteredGroupedMorphologyLocations",
    "ClusteredMorphologyLocations",
    "ClusteredPDSynapsesByCount",
    "ClusteredPDSynapsesByMaxDistance",
    "ClusteredPathDistanceMorphologyLocations",
    "ClusteredSynapsesByCount",
    "ClusteredSynapsesByMaxDistance",
    "CombinedNeuronSet",
    "ConnectivityMatrixExtraction",
    "ConnectivityMatrixExtractions",
    "ConstantCurrentClampSomaticStimulus",
    "CoupledScan",
    "EntityFromID",
    "ExcitatoryNeurons",
    "ExtracellularLocationSet",
    "ExtracellularLocationSetUnion",
    "FolderCompression",
    "FolderCompressions",
    "Form",
    "FormUnion",
    "FullySynchronousSpikeStimulus",
    "GridScan",
    "HyperpolarizingCurrentClampSomaticStimulus",
    "IDNeuronSet",
    "Info",
    "InhibitoryNeurons",
    "IntracellularLocationSet",
    "IntracellularLocationSetUnion",
    "LoadAssetMethod",
    "LinearCurrentClampSomaticStimulus",
    "MorphologyContainerization",
    "MorphologyContainerizationsForm",
    "MorphologyDecontainerization",
    "MorphologyDecontainerizationsForm",
    "MorphologyLocations",
    "MorphologyLocationsForm",
    "MorphologyMetrics",
    "MorphologyMetricsForm",
    "MorphologyMetricsOutput",
    "MultiBlockEntitySDKTest",
    "MultiBlockEntitySDKTestForm",
    "MultiPulseCurrentClampSomaticStimulus",
    "NamedPath",
    "NamedTuple",
    "NeuronPropertyFilter",
    "NeuronSet",
    "NeuronSetReference",
    "NeuronSetUnion",
    "NormallyDistributedCurrentClampSomaticStimulus",
    "OBIBaseModel",
    "OBIONE_Error",
    "PathDistanceConstrainedFractionOfSynapses",
    "PathDistanceConstrainedNumberOfSynapses",
    "PathDistanceWeightedFractionOfSynapses",
    "PathDistanceWeightedNumberOfSynapses",
    "RelativeNormallyDistributedCurrentClampSomaticStimulus",
    "PoissonSpikeStimulus",
    "PredefinedNeuronSet",
    "PropertyNeuronSet",
    "RandomGroupedMorphologyLocations",
    "RandomMorphologyLocations",
    "RandomlySelectedFractionOfSynapses",
    "RandomlySelectedNumberOfSynapses",
    "ReconstructionMorphologyFromID",
    "ReconstructionMorphologyValidation",
    "Recording",
    "RecordingReference",
    "RecordingUnion",
    "RegularTimestamps",
    "RelativeConstantCurrentClampSomaticStimulus",
    "RelativeLinearCurrentClampSomaticStimulus",
    "SectionIntracellularLocationSet",
    "Simulation",
    "SimulationsForm",
    "SimulationNeuronSetUnion",
    "SingleBlockEntitySDKTest",
    "SingleBlockEntityTestForm",
    "SingleBlockGenerateTest",
    "SingleBlockGenerateTestForm",
    "SingleCoordinateMixin",
    "SingleTimestamp",
    "SinusoidalCurrentClampSomaticStimulus",
    "SomaVoltageRecording",
    "TimeWindowSomaVoltageRecording",
    "StimulusReference",
    "StimulusUnion",
    "SubthresholdCurrentClampSomaticStimulus",
    "ScaleAcetylcholineUSESynapticManipulation",
    "SynapticMgManipulation",
    "nbS1POmInputs",
    "nbS1VPMInputs",
    "SynapseSetUnion",
    "Timestamps",
    "TimestampsReference",
    "TimestampsUnion",
    "Validation",
    "VolumetricCountNeuronSet",
    "VolumetricRadiusNeuronSet",
    "XYZExtracellularLocationSet",
    "db",
    "deserialize_obi_object_from_json_data",
    "deserialize_obi_object_from_json_file",
    "SimplexNeuronSet", 
    "SimplexMembershipBasedNeuronSet"
]

from obi_one.database.entity_from_id import (
    EntityFromID, LoadAssetMethod
)
from obi_one.database.reconstruction_morphology_from_id import (
    ReconstructionMorphologyFromID,
)

from obi_one.database.circuit_from_id import CircuitFromID

from obi_one.scientific.afferent_synapse_finder.specified_afferent_synapses_block import (
    AfferentSynapsesBlock,
    ClusteredPDSynapsesByCount,
    ClusteredPDSynapsesByMaxDistance,
    ClusteredSynapsesByCount,
    ClusteredSynapsesByMaxDistance,
    PathDistanceConstrainedFractionOfSynapses,
    PathDistanceConstrainedNumberOfSynapses,
    PathDistanceWeightedFractionOfSynapses,
    PathDistanceWeightedNumberOfSynapses,
    RandomlySelectedFractionOfSynapses,
    RandomlySelectedNumberOfSynapses,
)
from obi_one.scientific.basic_connectivity_plots.basic_connectivity_plots import (
    BasicConnectivityPlot,
    BasicConnectivityPlots,
)
from obi_one.scientific.circuit.circuit import Circuit
from obi_one.scientific.circuit.extracellular_location_sets import (
    ExtracellularLocationSet,
    XYZExtracellularLocationSet,
)
from obi_one.scientific.circuit.intracellular_location_sets import (
    IntracellularLocationSet,
    SectionIntracellularLocationSet,
)
from obi_one.scientific.circuit.neuron_sets import (
    CombinedNeuronSet,
    IDNeuronSet,
    NeuronPropertyFilter,
    NeuronSet,
    PredefinedNeuronSet,
    PropertyNeuronSet,
    VolumetricCountNeuronSet,
    VolumetricRadiusNeuronSet,
    SimplexNeuronSet,
    SimplexMembershipBasedNeuronSet,
    nbS1VPMInputs,
    nbS1POmInputs,
    rCA1CA3Inputs,
    AllNeurons,
    ExcitatoryNeurons,
    InhibitoryNeurons,
)
from obi_one.scientific.circuit_extraction.circuit_extraction import (
    CircuitExtraction,
    CircuitExtractions,
)
from obi_one.scientific.connectivity_matrix_extraction.connectivity_matrix_extraction import (
    ConnectivityMatrixExtraction,
    ConnectivityMatrixExtractions,
)
from obi_one.scientific.folder_compression.folder_compression import (
    FolderCompression,
    FolderCompressions,
)
from obi_one.scientific.morphology_containerization.morphology_containerization import (
    MorphologyContainerization,
    MorphologyContainerizationsForm,
)
from obi_one.scientific.morphology_containerization.morphology_decontainerization import (
    MorphologyDecontainerization,
    MorphologyDecontainerizationsForm,
)
from obi_one.scientific.morphology_locations.morphology_location_block import (
    ClusteredGroupedMorphologyLocations,
    ClusteredMorphologyLocations,
    ClusteredPathDistanceMorphologyLocations,
    RandomGroupedMorphologyLocations,
    RandomMorphologyLocations,
)
from obi_one.scientific.morphology_locations.morphology_location_form import (
    MorphologyLocations,
    MorphologyLocationsForm,
)
from obi_one.scientific.morphology_metrics.morphology_metrics import (
    MorphologyMetrics,
    MorphologyMetricsForm,
    MorphologyMetricsOutput,
)
from obi_one.scientific.simulation.recording import (
    Recording,
    SomaVoltageRecording,
    TimeWindowSomaVoltageRecording,
)
from obi_one.scientific.simulation.simulations import Simulation, SimulationsForm
from obi_one.scientific.simulation.stimulus import (
    ConstantCurrentClampSomaticStimulus,
    HyperpolarizingCurrentClampSomaticStimulus,
    LinearCurrentClampSomaticStimulus,
    MultiPulseCurrentClampSomaticStimulus,
    NormallyDistributedCurrentClampSomaticStimulus,
    RelativeNormallyDistributedCurrentClampSomaticStimulus,
    RelativeConstantCurrentClampSomaticStimulus,
    RelativeLinearCurrentClampSomaticStimulus,
    SinusoidalCurrentClampSomaticStimulus,
    SubthresholdCurrentClampSomaticStimulus,
    PoissonSpikeStimulus,
    FullySynchronousSpikeStimulus
)
from obi_one.scientific.simulation.timestamps import RegularTimestamps, Timestamps, SingleTimestamp
from obi_one.scientific.test_forms.test_form_single_block import (
    MultiBlockEntitySDKTest,
    MultiBlockEntitySDKTestForm,
    SingleBlockEntitySDKTest,
    SingleBlockEntityTestForm,
    SingleBlockGenerateTest,
    SingleBlockGenerateTestForm,
)
from obi_one.scientific.unions.unions_extracellular_location_sets import (
    ExtracellularLocationSetUnion,
)
from obi_one.scientific.unions.unions_form import (
    FormUnion,
)
from obi_one.scientific.unions.unions_intracellular_location_sets import (
    IntracellularLocationSetUnion,
)
from obi_one.scientific.unions.unions_neuron_sets import NeuronSetUnion, SimulationNeuronSetUnion, NeuronSetReference
from obi_one.scientific.unions.unions_recordings import RecordingUnion, RecordingReference
from obi_one.scientific.unions.unions_stimuli import StimulusUnion, StimulusReference
from obi_one.scientific.unions.unions_synapse_set import SynapseSetUnion
from obi_one.scientific.unions.unions_timestamps import TimestampsUnion, TimestampsReference

from obi_one.scientific.validations.reconstruction_morphology_validation import (
    ReconstructionMorphologyValidation,
)

from obi_one.scientific.unions.unions_manipulations import (
    SynapticMgManipulation,
    ScaleAcetylcholineUSESynapticManipulation,
)



import json
import os
from typing import ClassVar, Literal, Self, Annotated

from pydantic import Field, PrivateAttr, model_validator, NonNegativeInt, NonNegativeFloat, PositiveInt, PositiveFloat

from obi_one.core.block import Block
from obi_one.core.form import Form
from obi_one.core.single import SingleCoordinateMixin
from obi_one.core.info import Info
from obi_one.core.exception import OBIONE_Error
from obi_one.scientific.circuit.circuit import Circuit
from obi_one.scientific.circuit.neuron_sets import NeuronSet
from obi_one.scientific.unions.unions_extracellular_location_sets import (
    ExtracellularLocationSetUnion,
)
from obi_one.scientific.unions.unions_manipulations import SynapticManipulationsUnion, SynapticManipulationsReference
from obi_one.scientific.unions.unions_morphology_locations import MorphologyLocationUnion
from obi_one.scientific.unions.unions_neuron_sets import SimulationNeuronSetUnion, NeuronSetReference
from obi_one.scientific.unions.unions_recordings import RecordingUnion, RecordingReference
from obi_one.scientific.unions.unions_stimuli import StimulusUnion, StimulusReference
from obi_one.scientific.unions.unions_synapse_set import SynapseSetUnion
from obi_one.scientific.unions.unions_timestamps import TimestampsUnion, TimestampsReference

from obi_one.database.circuit_from_id import CircuitFromID

import entitysdk
from collections import OrderedDict

from datetime import UTC, datetime

from pathlib import Path

import logging
L = logging.getLogger(__name__)

from enum import StrEnum
class BlockGroup(StrEnum):
    """Authentication and authorization errors."""

    SETUP_BLOCK_GROUP = "Setup"
    STIMULI_RECORDINGS_BLOCK_GROUP = "Stimuli & Recordings"
    CIRUIT_COMPONENTS_BLOCK_GROUP = "Circuit Components"
    EVENTS_GROUP = "Events"
    CIRCUIT_MANIPULATIONS_GROUP = "Circuit Manipulations"
    

class SimulationsForm(Form):
    """Simulations Form."""

    single_coord_class_name: ClassVar[str] = "Simulation"
    name: ClassVar[str] = "Simulation Campaign"
    description: ClassVar[str] = "SONATA simulation campaign"

    class Config:
        json_schema_extra = {
            "block_block_group_order": [BlockGroup.SETUP_BLOCK_GROUP,
                                     BlockGroup.STIMULI_RECORDINGS_BLOCK_GROUP,
                                     BlockGroup.CIRUIT_COMPONENTS_BLOCK_GROUP,
                                     BlockGroup.EVENTS_GROUP,
                                     BlockGroup.CIRCUIT_MANIPULATIONS_GROUP]
        }

    timestamps: dict[str, TimestampsUnion] = Field(default_factory=dict, title="Timestamps", reference_type=TimestampsReference.__name__, description="Timestamps for the simulation.", singular_name="Timestamps", group=BlockGroup.SETUP_BLOCK_GROUP, group_order=0)
    stimuli: dict[str, StimulusUnion] = Field(default_factory=dict, title="Stimuli", reference_type=StimulusReference.__name__, description="Stimuli for the simulation.", singular_name="Stimulus", group=BlockGroup.STIMULI_RECORDINGS_BLOCK_GROUP, group_order=0)
    recordings: dict[str, RecordingUnion] = Field(default_factory=dict, reference_type=RecordingReference.__name__, description="Recordings for the simulation.", singular_name="Recording", group=BlockGroup.STIMULI_RECORDINGS_BLOCK_GROUP, group_order=1)
    neuron_sets: dict[str, SimulationNeuronSetUnion] = Field(default_factory=dict, reference_type=NeuronSetReference.__name__, description="Neuron sets for the simulation.", singular_name="Neuron Set", group=BlockGroup.CIRUIT_COMPONENTS_BLOCK_GROUP, group_order=0)
    synaptic_manipulations: dict[str, SynapticManipulationsUnion] = Field(default_factory=dict, reference_type=SynapticManipulationsReference.__name__, description="Synaptic manipulations for the simulation.", singular_name="Synaptic Manipulation", group=BlockGroup.CIRUIT_COMPONENTS_BLOCK_GROUP, group_order=1)

    class Initialize(Block):
        circuit: list[Circuit] | Circuit | CircuitFromID | list[CircuitFromID]
        node_set: Annotated[NeuronSetReference, Field(title="Neuron Set", description="Neuron set to simulate.")]
        simulation_length: list[NonNegativeFloat] | NonNegativeFloat = Field(default=1000.0, title="Duration", description="Simulation length in milliseconds (ms).", units="ms")
        extracellular_calcium_concentration: list[NonNegativeFloat] | NonNegativeFloat = Field(default=1.1, title="Extracellular Calcium Concentration", description="Extracellular calcium concentration around the synapse in millimoles (mM). Increasing this value increases the probability of synaptic vesicle release, which in turn increases the level of network activity. In vivo values are estimated to be ~0.9-1.2mM, whilst in vitro values are on the order of 2mM.", units="mM")
        v_init: list[float] | float = Field(default=-80.0, title="Initial Voltage", description="Initial membrane potential in millivolts (mV).", units="mV")
        random_seed: list[int] | int = Field(default=1, description="Random seed for the simulation.")
        
        _spike_location: Literal["AIS", "soma"] | list[Literal["AIS", "soma"]] = PrivateAttr(default="soma")
        _sonata_version: list[NonNegativeFloat] | NonNegativeFloat = PrivateAttr(default=2.4) 
        _target_simulator: Literal["NEURON", "CORENEURON"] | list[Literal["NEURON", "CORENEURON"]] = PrivateAttr(default="NEURON") # Target simulator for the simulation
        _timestep: list[PositiveFloat] | PositiveFloat = PrivateAttr(default=0.025) # Simulation time step in ms

    initialize: Initialize = Field(title="Initialization", description="Parameters for initializing the simulation.", group=BlockGroup.SETUP_BLOCK_GROUP, group_order=1)
    info: Info = Field(title="Info", description="Information about the simulation campaign.", group=BlockGroup.SETUP_BLOCK_GROUP, group_order=0)


    def initialize_db_campaign(self, output_root: Path, multiple_value_parameters_dictionary={}, db_client: entitysdk.client.Client=None):

        """Initializes the simulation campaign in the database."""
        L.info("1. Initializing simulation campaign in the database...")

        L.info(f"-- Register SimulationCampaign Entity")
        self._campaign = db_client.register_entity(
            entitysdk.models.SimulationCampaign(
                name=self.info.campaign_name,
                description=self.info.campaign_description,
                entity_id=self.initialize.circuit.id_str if isinstance(self.initialize.circuit, CircuitFromID) else self.initialize.circuit[0].id_str,
                scan_parameters=multiple_value_parameters_dictionary,
            )
        )

        L.info(f"-- Upload campaign_generation_config")
        _ = db_client.upload_file(
            entity_id=self._campaign.id,
            entity_type=entitysdk.models.SimulationCampaign,
            file_path=output_root / "run_scan_config.json",
            file_content_type="application/json",
            asset_label='campaign_generation_config'
        )

        # L.info(f"-- Upload campaign_summary")
        # _ = db_client.upload_file(
        #     entity_id=self._campaign.id,
        #     entity_type=entitysdk.models.SimulationCampaign,
        #     file_path=Path(output_root, "bbp_workflow_campaign_config.json"),
        #     file_content_type="application/json",
        #     asset_label='campaign_summary'
        # )

        return self._campaign
    
    def save(self, simulations, db_client: entitysdk.client.Client) -> None:

        L.info("3. Saving completed simulation campaign generation")

        L.info(f"-- Register SimulationGeneration Entity")
        db_client.register_entity(
            entitysdk.models.SimulationGeneration(
                start_time=datetime.now(UTC),
                used=[self._campaign],
                generated=simulations,
            )
        )

        return None

    
    def add(self, block: Block, name:str='') -> None:

        block_dict_name = self.block_mapping[block.__class__.__name__]["block_dict_name"]
        reference_type_name = self.block_mapping[block.__class__.__name__]["reference_type"]

        if name in self.__dict__.get(block_dict_name).keys():
            raise OBIONE_Error(f"Block with name '{name}' already exists in '{block_dict_name}'!")
        
        else: 
            reference_type = globals()[reference_type_name]
            ref = reference_type(block_dict_name=block_dict_name, block_name=name)
            block.set_ref(ref)
            self.__dict__[block_dict_name][name] = block


    def set(self, block: Block, name: str = '') :
        """Sets a block in the form."""
        self.__dict__[name] = block


    
    # Below are initializations of the individual components as part of a simulation
    # by setting their simulation_level_name as the one used in the simulation form/GUI
    # TODO: Ensure in GUI that these names don't have spaces or special characters
    @model_validator(mode="after")
    def initialize_timestamps(self) -> Self:
        """Initializes timestamps within simulation campaign."""
        for _k, _v in self.timestamps.items():
            _v.set_simulation_level_name(_k)
        return self

    @model_validator(mode="after")
    def initialize_stimuli(self) -> Self:
        """Initializes stimuli within simulation campaign."""
        for _k, _v in self.stimuli.items():
            _v.set_simulation_level_name(_k)
        return self

    @model_validator(mode="after")
    def initialize_recordings(self) -> Self:
        """Initializes recordings within simulation campaign."""
        for _k, _v in self.recordings.items():
            _v.set_simulation_level_name(_k)
        return self

    @model_validator(mode="after")
    def initialize_neuron_sets(self) -> Self:
        """Initializes neuron sets within simulation campaign."""
        for _k, _v in self.neuron_sets.items():
            _v.set_simulation_level_name(_k)
        return self

    @model_validator(mode="after")
    def initialize_synaptic_manipulations(self) -> Self:
        """Initializes manipulationms within simulation campaign."""
        for _k, _v in self.synaptic_manipulations.items():
            _v.set_simulation_level_name(_k)
        return self


# def _resolve_neuron_set_dictionary(neuron_set, _circuit):
#         """Resolves a neuron set based on current coordinate circuit's default node population and \
#             returns its dictionary.
#         """
#         nset_def = neuron_set.get_node_set_definition(
#             _circuit, _circuit.default_population_name
#         )
#         # FIXME: Better handling of (default) node population in case there is more than one
#         # FIXME: Inconsistency possible in case a node set definition would span multiple populations
#         #        May consider force_resolve_ids=False to enforce resolving into given population
#         #        (but which won't be a human-readable representation any more)
#         name = neuron_set.name
#         dictionary = {name: nset_def}
#         return name, dictionary


class Simulation(SimulationsForm, SingleCoordinateMixin):
    """Only allows single values and ensures nested attributes follow the same rule."""

    CONFIG_FILE_NAME: ClassVar[str] = "simulation_config.json"
    NODE_SETS_FILE_NAME: ClassVar[str] = "node_sets.json"

    _sonata_config: dict = PrivateAttr(default={})

    def generate(self, db_client: entitysdk.client.Client = None):
        """Generates SONATA simulation config .json file."""

        # Set _circuit parameter based on the type of initialize.circuit
        # _circuit is used through-out generate rather than self.initialize.circuit
        _circuit = None
        if isinstance(self.initialize.circuit, Circuit):
            L.info("initialize.circuit is a Circuit instance.")
            _circuit = self.initialize.circuit
            self._sonata_config["network"] = self.initialize.circuit.path

        if isinstance(self.initialize.circuit, CircuitFromID):
            L.info("initialize.circuit is a CircuitFromID instance.")
            self._circuit_id = self.initialize.circuit.id_str

            for asset in self.initialize.circuit.entity(db_client=db_client).assets:
                if asset.label == "sonata_circuit":
                    self.initialize.circuit.download_circuit_directory(dest_dir=self.coordinate_output_root, db_client=db_client)
                    _circuit = Circuit(name=self.initialize.circuit.entity(db_client=db_client).name, path=str(self.coordinate_output_root / asset.path / "circuit_config.json"))
                    self._sonata_config["network"] = asset.path + "/" + Path(_circuit.path).name
                    break


        self._sonata_config["output"] = {
            "output_dir": "output",
            "spikes_file": "spikes.h5"
        }

        self._sonata_config["version"] = self.initialize._sonata_version
        self._sonata_config["target_simulator"] = self.initialize._target_simulator

        self._sonata_config["run"] = {}
        self._sonata_config["run"]["dt"] = self.initialize._timestep
        self._sonata_config["run"]["random_seed"] = self.initialize.random_seed
        self._sonata_config["run"]["tstop"] = self.initialize.simulation_length

        self._sonata_config["conditions"] = {}
        self._sonata_config["conditions"]["extracellular_calcium"] = (
            self.initialize.extracellular_calcium_concentration
        )
        self._sonata_config["conditions"]["v_init"] = self.initialize.v_init
        self._sonata_config["conditions"]["spike_location"] = self.initialize._spike_location

        self._sonata_config["conditions"]["mechanisms"] = {
            "ProbAMPANMDA_EMS": {"init_depleted": True, "minis_single_vesicle": True},
            "ProbGABAAB_EMS": {"init_depleted": True, "minis_single_vesicle": True},
        }

        # Generate stimulus input configs
        self._sonata_config["inputs"] = {}
        for stimulus_key, stimulus in self.stimuli.items():
            if hasattr (stimulus, "generate_spikes"):
                stimulus.generate_spikes(_circuit, self.coordinate_output_root, self.initialize.simulation_length, source_node_population=_circuit.default_population_name)
            self._sonata_config["inputs"].update(stimulus.config(_circuit, _circuit.default_population_name))
            # 

        # Generate recording configs
        self._sonata_config["reports"] = {}
        for recording_key, recording in self.recordings.items():
            self._sonata_config["reports"].update(recording.config(_circuit, _circuit.default_population_name, self.initialize.simulation_length))

        # Generate list of synaptic manipulation configs (executed in the order in the list)
        # FIXME: Check and make sure that the order in the self.synaptic_manipulations dict is preserved!
        manipulation_list = []
        for manipulation_key, manipulation in self.synaptic_manipulations.items():
            manipulation_list.append(manipulation.config())
        if len(manipulation_list) > 0:
            self._sonata_config["connection_overrides"] = manipulation_list

        # Resolve neuron sets and add them to the SONATA circuit object
        # NOTE: The name that is used as neuron_sets dict key is always used as name for a new node
        # set, even for a PredefinedNeuronSet in which case a new node set will be created which just
        # references the existing one. This is the most consistent behavior since it will behave
        # exactly the same no matter if random subsampling is used or not. But this also means that
        # existing names cannot be used as dict keys.
        os.makedirs(self.coordinate_output_root, exist_ok=True)
        c = _circuit.sonata_circuit
        for _name, _nset in self.neuron_sets.items():
            # Resolve node set based on current coordinate circuit's default node population
            # FIXME: Better handling of (default) node population in case there is more than one
            # FIXME: Inconsistency possible in case a node set definition would span multiple populations
            #        May consider force_resolve_ids=False to enforce resolving into given population
            #        (but which won't be a human-readable representation any more)
            assert _name == _nset.name, (
                "Neuron set name mismatch!"
            )  # This should never happen if properly initialized

            if self.initialize.node_set.block.name == _name:
                assert self._sonata_config.get("node_set") is None, (
                    "Node set config entry already defined!"
                )

                # Assert that simulation neuron set is biophysical
                if _nset.population_type(_circuit, _circuit.default_population_name) != "biophysical":
                    raise OBIONE_Error(
                        f"Simulation Neuron Set (Initialize -> Neuron Set): '{_name}' is not biophysical!"
                    )

                self._sonata_config["node_set"] = _name

            # Add node set to SONATA circuit object
            # (will raise an error in case already existing)
            nset_def = _nset.get_node_set_definition(
                _circuit, _circuit.default_population_name, force_resolve_ids=True
            )
            NeuronSet.add_node_set_to_circuit(c, {_name: nset_def}, overwrite_if_exists=False)

        # Write node sets from SONATA circuit object to .json file
        # (will raise an error if file already exists)
        NeuronSet.write_circuit_node_set_file(
            c,
            self.coordinate_output_root,
            file_name=self.NODE_SETS_FILE_NAME,
            overwrite_if_exists=False,
        )
        self._sonata_config["node_sets_file"] = self.NODE_SETS_FILE_NAME

        # Write simulation config file (.json)
        simulation_config_path = os.path.join(self.coordinate_output_root, self.CONFIG_FILE_NAME)
        with open(simulation_config_path, "w") as f:
            json.dump(self._sonata_config, f, indent=2)


    def save(self, campaign: entitysdk.models.SimulationCampaign, db_client: entitysdk.client.Client) -> None:
        """Saves the simulation to the database."""
        
        L.info(f"2.{self.idx} Saving simulation {self.idx} to database...")
        
        L.info(f"-- Register Simulation Entity")
        simulation = db_client.register_entity(
            entitysdk.models.Simulation(
                name=f"sim-{self.idx}",
                description=f"sim-{self.idx}",
                scan_parameters=self.single_coordinate_scan_params.dictionary_representaiton(),
                entity_id=self._circuit_id,
                simulation_campaign_id=campaign.id,
                
            )
        )

        L.info(f"-- Upload simulation_generation_config")
        _ = db_client.upload_file(
            entity_id=simulation.id,
            entity_type=entitysdk.models.Simulation,
            file_path=Path(self.coordinate_output_root, "run_coordinate_instance.json"),
            file_content_type="application/json",
            asset_label='simulation_generation_config'
        )

        L.info(f"-- Upload sonata_simulation_config")
        _ = db_client.upload_file(
            entity_id=simulation.id,
            entity_type=entitysdk.models.Simulation,
            file_path=Path(self.coordinate_output_root, "simulation_config.json"),
            file_content_type="application/json",
            asset_label='sonata_simulation_config'
        )
        
        L.info(f"-- Upload custom_node_sets")
        _ = db_client.upload_file(
            entity_id=simulation.id,
            entity_type=entitysdk.models.Simulation,
            file_path=Path(self.coordinate_output_root, "node_sets.json"),
            file_content_type="application/json",
            asset_label='custom_node_sets'
        )

        L.info(f"-- Upload spike replay files")
        for input in self._sonata_config["inputs"]:
            if "spike_file" in list(self._sonata_config["inputs"][input]):
                spike_file = self._sonata_config["inputs"][input]["spike_file"]
                if spike_file is not None:
                    _ = db_client.upload_file(
                        entity_id=simulation.id,
                        entity_type=entitysdk.models.Simulation,
                        file_path=Path(self.coordinate_output_root, spike_file),
                        file_content_type="application/x-hdf5",
                        asset_label='replay_spikes'
                    )

        return simulation


from pathlib import Path

from entitysdk.client import Client
from entitysdk.common import ProjectContext
from obi_auth import get_token

class FixedTokenManager:
    """A fixed token manager that always returns the same token."""

    def __init__(self, token: str):
        self._token = token

    def get_token(self) -> str:
        return self._token

class DatabaseManager:
    def __init__(self) -> None:
        """Initialize the DatabaseManager with default values."""
        self.client = None
        self.token = None
        self.entity_file_store_path = None

    def initialize(
        self,
        virtual_lab_id: str,
        project_id: str,
        entity_file_store_root: Path = Path("../../obi-output"),
        entitycore_api_url: str | None = None, # If None, it will use the default from settings.ENTITYCORE_URL
    ) -> None:
        """Initialize the database connection and set up the file store path."""
        self.entity_file_store_path = entity_file_store_root / "obi-entity-file-store"

        self.entity_file_store_path.mkdir(parents=True, exist_ok=True)

        # Staging
        self.token = get_token(environment="staging")
        project_context = ProjectContext(virtual_lab_id=virtual_lab_id, project_id=project_id)

        # self.token.credentials
        token_manager = FixedTokenManager(self.token)

        if entitycore_api_url is None:
            entitycore_api_url = settings.ENTITYCORE_URL

        self.client = Client(
            api_url=entitycore_api_url,
            project_context=project_context,
            token_manager=token_manager,
        )

        """
        Local. Not fully working
        project_context = ProjectContext(virtual_lab_id=virtual_lab_id, project_id=project_id)
        self.client = Client(api_url=entitycore_api_url, project_context=project_context)
        self.token = os.getenv("ACCESS_TOKEN", "XXX")
        """


db = DatabaseManager()



import warnings

import morphio
import numpy as np
import pandas as pd
from scipy import stats

try:
    from conntility.subcellular import MorphologyPathDistanceCalculator
except ImportError:
    warnings.warn("Connectome functionalities not available", UserWarning, stacklevel=1)

_SEC_ID = "section_id"
_SEG_ID = "segment_id"
_SEG_OFF = "segment_offset"
_SEG_LEN = "segment_length"
_SEC_TYP = "section_type"
_SEC_LOC = "normalized_section_offset"
_SOM_PAD = "path_distance"
_CEN_IDX = "reference_loc_index"
_SEG_MIN = "min_seg_offset"
_SEG_MAX = "max_seg_offset"
_PRE_IDX = "source_index"


def path_distance_all_segments_from(
    locs_ref: pd.DataFrame,
    m: morphio.Morphology,
    path_distance_calculator: MorphologyPathDistanceCalculator,
    normalized_seg_loc: float = 0.5,
    lst_sec_types: list | None = None,
):
    """Path distances from reference locations to all segments of a morphology.
    
    Calculates path distances from reference locations on a morphology to all segments of the
    same morphology.

    Args:
        locs_ref (pd.DataFrame): Defines the reference locations. For details how, see
        conntility.subcellular.MorphologyPathDistanceCalculator.
        m (morphio.Morphology): Morphology to use.
        path_distance_calculator (conntility.subcellular.MorphologyPathDistanceCalculator): \
            Path distance calculator
        must be created on morphology m.
        normalized_seg_loc (float, default: 0.5): Between 0 and 1. Normalized segment location.
        Path distances will be calculated to the part of each segment.
        lst_sec_types (list, default=None): List of section types to consider.
        If None, all types are considered.
    """
    kwargs = {"str_section_id": _SEC_ID, "str_segment_id": _SEG_ID, "str_offset": _SEG_OFF}
    locs_all = pd.concat(
        [
            pd.DataFrame(
                {
                    _SEG_ID: np.arange(m.sections[i].n_points - 1),
                    _SEC_ID: np.ones(m.sections[i].n_points - 1, dtype=int) * (i + 1),
                    _SEC_TYP: np.ones(m.sections[i].n_points - 1, dtype=int)
                    * int(m.sections[i].type),
                    _SEG_LEN: np.linalg.norm(np.diff(m.sections[i].points, axis=0), axis=1),
                }
            )
            for i in range(len(m.sections))
        ],
        axis=0,
    ).reset_index(drop=True)
    if lst_sec_types is not None:
        locs_all = locs_all.loc[locs_all[_SEC_TYP].isin(lst_sec_types)].reset_index(drop=True)
    locs_all[_SEG_OFF] = normalized_seg_loc * locs_all[_SEG_LEN]

    path_distances = path_distance_calculator.path_distances(locs_ref, locs_all, **kwargs)
    path_distances = [
        pd.Series(_path_distances, name=_SOM_PAD) for _path_distances in path_distances
    ]

    return pd.concat(
        [pd.concat([locs_all, _path_distances], axis=1) for _path_distances in path_distances],
        axis=0,
        keys=range(len(path_distances)),
        names=[_CEN_IDX],
    )


def select_segments_as_cluster_centers(n_pick, locs, distr, lst_sec_types: list = None):
    p = distr.pdf(locs[_SOM_PAD])
    if lst_sec_types is not None:
        p = p * locs[_SEC_TYP].isin(lst_sec_types).astype(float)

    selected_ids = np.random.choice(locs.index, n_pick, p=p / p.sum())
    return locs.iloc[selected_ids]


def find_normalized_interval_below_zero(a: float, b: float) -> tuple[float, float]:
    """Find sub-interval of [0, 1] where f is < 0.

    Being given a and b, and assuming f(0) = a, f(1) = b, this function calculates the
    sub-interval of [0, 1] where f is < 0.

    Args:
        a (float): f(0)
        b (float): f(1)
    """
    min_norm_path_distances = float(0)
    max_norm_path_distances = float(1)
    if (b - a) > 0:  # becomes invalid near the end
        max_norm_path_distances = np.minimum(-a / (b - a), max_norm_path_distances)
    elif (b - a) == 0:
        max_norm_path_distances = 0.0
    else:
        min_norm_path_distances = np.maximum(-a / (b - a), min_norm_path_distances)
    return min_norm_path_distances, max_norm_path_distances


def min_max_offset_in_segment(row: pd.Series, max_distance: float) -> pd.Series:
    """Determines offsets from segment starting point closer than max_distance. \
    
    Determines which offsets from the starting point of a segment are closer in path distance \
        than a specified maximum value.

    Args:
        row (pd.Series): Must specify path distance at the start and end of the segment,
        plus the length of the segment. For the names of the corresponding entries check the code.
        max_distance (float): Maximum path distance allowed.
    """
    a = row[_SOM_PAD + "_start"] - max_distance
    b = row[_SOM_PAD + "_end"] - max_distance
    l = row[_SEG_LEN]
    min_norm_path_distances, max_norm_path_distances = find_normalized_interval_below_zero(a, b)

    return pd.Series({_SEG_MIN: min_norm_path_distances * l, _SEG_MAX: max_norm_path_distances * l})


def min_max_offset_for_center_segment(row: pd.Series, max_distance: float) -> pd.Series:
    """Determines offsets from segment starting point closer than max_distance. \
    
    Determines which offsets from the starting point of a segment are closer in path distance \
        than a specified maximum value. Specialized version to be used for the segment that \
        contains the point that path distances are relative to.

    Args:
        row (pd.Series): Must specify path distance at the start and end of the segment,
        plus the length of the segment. For the names of the corresponding entries check the code.
        max_distance (float): Maximum path distance allowed.
    """
    a = row[_SOM_PAD + "_start"] - max_distance
    b = row[_SOM_PAD + "_end"] - max_distance
    l = row[_SEG_LEN]
    c = -max_distance
    min_norm_path_distances, _ = find_normalized_interval_below_zero(a, c)
    _, max_norm_path_distances = find_normalized_interval_below_zero(c, b)

    return pd.Series(
        {
            _SEG_MIN: 0.5 * min_norm_path_distances * l,
            _SEG_MAX: (0.5 + 0.5 * max_norm_path_distances) * l,
        }
    )


def candidate_segments_all_morphology(locs) -> pd.DataFrame:
    """Creates a dataframe matching the format of 'candidate_segments_for_center' (*).

    * But where all parts of the morphology are admitted as candidates. \
            For details, see that function.
    """
    return pd.DataFrame(
        {_SEG_MIN: np.zeros(len(locs), dtype=float), _SEG_MAX: locs[_SEG_LEN].to_numpy()},
        index=locs.index,
    )


def candidate_segments_for_center(
    ids_center: list,
    locs: pd.DataFrame,
    max_dist: float,
    m: morphio.Morphology,
    path_distance_calculator: MorphologyPathDistanceCalculator,
    lst_sec_types: list | None = None,
) -> pd.DataFrame:
    """Creates dataframe specifying candidate locations to be picked around center of a cluster.

    That is, finds for all segments intervals that are closer than a maximum distance to the \
        center. If no part of a segment is closer than the value, then the segment is omitted \
            from the output.

    Output is a DataFrame with one row per candidate segment. Columns specify the minimum and \
        maximum within-segment offset in um that is still within the maximum distance.

    Args:
        ids_center (list): List of integers that can be used as index into 'locs'. Specifies the \
            locations of the centers picked.
        locs (pd.DataFrame): dataframe of section id, segment id, soma path distance of all \
            segments
        max_dist (float): Maximum distance from the center allowed
        m (morphio.morphology): morphology to use
        path_distance_calculator (conntility.subcellular.MorphologyPathDistanceCalculator): must \
            be defined on 'm'
        lst_sec_types (list, default=None): List of section types allowed. If None, then all \
            types are allowed.
    """
    loc_center = locs.loc[ids_center]
    res_seg_start = path_distance_all_segments_from(
        loc_center, m, path_distance_calculator, 0.0, lst_sec_types=lst_sec_types
    ).rename(columns={_SOM_PAD: _SOM_PAD + "_start"})
    res_seg_end = path_distance_all_segments_from(
        loc_center, m, path_distance_calculator, 1.0, lst_sec_types=lst_sec_types
    ).rename(columns={_SOM_PAD: _SOM_PAD + "_end"})

    res = pd.concat([res_seg_start, res_seg_end[[_SOM_PAD + "_end"]]], axis=1)

    valid = (res[_SOM_PAD + "_start"] < max_dist) | (res[_SOM_PAD + "_end"] < max_dist)
    valid[list(enumerate(ids_center))] = True
    cands = res.loc[valid]

    intervals = cands.apply(min_max_offset_in_segment, args=(max_dist,), axis=1)
    for tpl in enumerate(ids_center):
        intervals.loc[tpl] = min_max_offset_for_center_segment(res.loc[tpl], max_dist)
    return intervals


def select_places_from_candidate_list(
    n: int, cands: pd.DataFrame, locs: pd.DataFrame
) -> pd.DataFrame:
    """From a list of candidate segments, pick randomly the specified number of locations.

    Locations are identified not just by their segment, but also the offset within the segment.
    Each candidate location is equally likely (no bias).

    Args:
        n (int): Number of locations to pick

        cands (pd.DataFrame): dataframe specifying segments and intervals on the segments that
        are valid to be picked. Output of 'candidate_segments_for_center'.

        locs (pd.DataFrame): dataframe of all segments on the morphology. Its index must be
        consistent with the index of 'cands'.
    """

    def randomly_pick_with_p(dataframe_in: pd.DataFrame) -> pd.DataFrame:
        dataframe_in = dataframe_in.droplevel(0)
        selected = dataframe_in.loc[
            np.random.choice(dataframe_in.index, n, p=dataframe_in["p"] / dataframe_in["p"].sum())
        ]
        selected = selected[_SEG_MIN] + np.random.rand(n) * (
            selected[_SEG_MAX] - selected[_SEG_MIN]
        )

        output = locs.loc[selected.index].sort_index().drop(columns=[_SEG_OFF])
        output[_SEG_OFF] = selected.values
        return output

    p = cands.diff(axis=1).values[:, 1]
    p[np.isnan(p)] = 0
    cands["p"] = p

    return cands.groupby(_CEN_IDX).apply(randomly_pick_with_p)


def map_presynaptic_ids(dataframe: pd.DataFrame, n_per_center: int) -> None:
    """Assign ids to generated locations.

    Locations around the same center are split into the specified number of groups and assigned \
        unique ids.

    Does not return an output, instead adds a column to the input dataframe.

    Args:
        dataframe (pd.DataFrame): Contains the selected locations. Defined by section id,
        segment id, segment offset, etc.

        n_per_center (int): Numner of unique identifiers per center.
    """
    rnd = np.random.randint(0, n_per_center, len(dataframe))
    dataframe[_PRE_IDX] = dataframe[_CEN_IDX] * n_per_center + rnd


def add_normalized_section_offset(
    dataframe: pd.DataFrame,
    m: morphio.Morphology,
    path_distance_calculator: MorphologyPathDistanceCalculator,
) -> None:
    """Adds the normalized section offset to a dataframe that defines morphology locations.

    Args:
        dataframe (pd.DataFrame): Contains the selected locations. Defined by section id,
        segment id, segment offset, etc.

        m (morphio.Morphology): morphology used

        path_distance_calculator (conntility.subcellular.MorphologyPathDistanceCalculator): Must \
            be defined on m.
    """
    sec_lengths = np.array(
        [np.linalg.norm(np.diff(sec.points, axis=0), axis=1).sum() for sec in m.sections]
    )
    sec_o = path_distance_calculator.O[dataframe[_SEC_ID] - 1, dataframe[_SEG_ID]]
    sec_l = sec_lengths[dataframe[_SEC_ID] - 1]
    dataframe[_SEC_LOC] = sec_o / sec_l


def generate_neurite_locations_on(
    m: morphio.Morphology,
    n_centers: int,
    n_per_center: int,
    srcs_per_center: int,
    center_path_distances_mean: float,
    center_path_distances_sd: float,
    max_dist_from_center: float,
    lst_section_types: list | None = None,
    seed: float | None = None,
) -> pd.DataFrame:
    """Generates a specified number of morphology locations according to complex specifications.

    Locations are defined along morphology skeleton, i.e. they are 1d locations, not membrane \
        locations.

    Locations are 'clustered' around center locations on the morphology. If that is not required, \
        the parameter 'max_dist_from_center' can be set to None and clustering is ignored.

    Otherwise, the locations of the centers can be parameterized by specifying the mean and \
        standard deviation of a Gaussian. The probability that any given neurite location is \
        picked is then proportional to the value of the Gaussian at the path distance from \
        the soma of the location. Note that this is *not* equivalent to saying that the \
        distribution of path distances of the output locations follows the Gaussian. \
        This is because some path distances are more common than others on the input morphology.

    If you do not want to constrain path distances, simply set the standard deviation of the \
        Gaussian to a very high value and then all morphology locations are equally likely.

    Output is a pd.DataFrame with one row per location. Columns define:
      - segment id
      - section id
      - section type
      - segment offset
      - soma path distance
      - normalized section offset
      - center id
      - source id

    where "center id" is an identifier for the center that a location is associated with. Even if
    'max_dist_from_center' is set to None, center ids are still generated.

    "source id" is defining a conceptual grouping that is generated according to specifications.\
        For example, when using this function to generate a synaptome, this could be the \
        presynaptic neuron id.

    Args:
        m (morphio.Morphology): The morphology to use.

        n_centers (int): Number of centers to group locations around. If not needed, set to 1.

        n_per_center (int): Number of locations to generate for each center.

        srcs_per_center (int): Number of source ids per center. Used to generate the source id.
        If not needed, set to 1.

        center_path_distances_mean (float): Mean of the Gaussian determining center path distances

        center_path_distances_sd (float): Standard deviation of the same Gaussian. If path \
            distances are no concern, then set this to 1E20 or a similarly large number

        max_dist_from_center (float): Maximum path distance from the center for the generated\
            locations.

        lst_section_types (list, default=None): List of section types that are valid targets. If \
            None, all types are valid.

        seed (optional): Random seed. For reproducability.
    """
    if seed is not None:
        np.random.seed(seed)

    path_distance_calculator = MorphologyPathDistanceCalculator(m)

    soma = pd.DataFrame({_SEC_ID: [0], _SEG_ID: [0], _SEG_OFF: [0]})

    locs = path_distance_all_segments_from(
        soma, m, path_distance_calculator, lst_sec_types=lst_section_types
    ).loc[0]

    if (
        max_dist_from_center is None
    ):  # Simpler case. If clustering around a center is not a concern.
        cands_all_morph = candidate_segments_all_morphology(locs)
        lst_candidates_per_center = pd.concat(
            [cands_all_morph for _ in range(n_centers)],
            axis=0,
            keys=range(n_centers),
            names=[_CEN_IDX],
        )
    else:
        distr = stats.norm(center_path_distances_mean, center_path_distances_sd)
        centers = select_segments_as_cluster_centers(n_centers, locs, distr, lst_sec_types=None)
        lst_candidates_per_center = candidate_segments_for_center(
            centers.index,
            locs,
            max_dist_from_center,
            m,
            path_distance_calculator,
            lst_sec_types=lst_section_types,
        )

    all_clusters = select_places_from_candidate_list(n_per_center, lst_candidates_per_center, locs)

    # Which columns do you want in the output?
    relevant_cols = [_SEG_ID, _SEC_ID, _SEC_TYP, _SEG_OFF, _SOM_PAD]
    all_clusters = all_clusters.droplevel(1)[relevant_cols].reset_index()

    map_presynaptic_ids(all_clusters, srcs_per_center)
    add_normalized_section_offset(all_clusters, m, path_distance_calculator)
    return all_clusters



import abc
from typing import Self

import morphio
from pydantic import Field, model_validator

from obi_one.core.block import Block

from .specified_morphology_locations import _CEN_IDX, generate_neurite_locations_on


class MorphologyLocationsBlock(Block, abc.ABC):
    """Base class representing parameterized locations on morphology skeletons."""

    random_seed: int | list[int] = Field(
        default=0, name="Random seed", description="Seed for the random generation of locations"
    )
    number_of_locations: int | list[int] = Field(
        default=1,
        name="Number of locations",
        description="Number of locations to generate on morphology",
    )
    section_types: tuple[int, ...] | list[tuple[int, ...]] | None = Field(
        default=None,
        name="Section types",
        description="Types of sections to generate locations on. 2: axon, 3: basal, 4: apical",
    )

    @abc.abstractmethod
    def _make_points(self, morphology: morphio.Morphology):
        """Returns a generated list of points for the morphology."""

    @abc.abstractmethod
    def _check_parameter_values(self):
        """Do specific checks on the validity of parameters."""

    @model_validator(mode="after")
    def check_parameter_values(self) -> Self:
        # Only check whenever list are resolved to individual objects
        self._check_parameter_values()
        return self

    def points_on(self, morphology: morphio.Morphology):
        self.enforce_no_lists()
        return self._make_points(morphology)


class RandomMorphologyLocations(MorphologyLocationsBlock):
    """Completely random locations without constraint."""

    def _make_points(self, morphology: morphio.Morphology):
        locs = generate_neurite_locations_on(
            morphology,
            n_centers=1,
            n_per_center=self.number_of_locations,
            srcs_per_center=1,
            center_path_distances_mean=0.0,
            center_path_distances_sd=0.0,
            max_dist_from_center=None,
            lst_section_types=self.section_types,
            seed=self.random_seed,
        ).drop(columns=[_CEN_IDX])
        return locs

    def _check_parameter_values(self):
        # Only check whenever list are resolved to individual objects
        if not isinstance(self.number_of_locations, list):
            assert self.number_of_locations > 0, "Number of locations must be at least one!"


class RandomGroupedMorphologyLocations(MorphologyLocationsBlock):
    """Completely random locations, but grouped into abstract groups."""

    n_groups: int | list[int] = Field(
        default=1,
        name="Number of groups",
        description="Number of groups of locations to \
            generate",
    )

    def _make_points(self, morphology: morphio.Morphology):
        locs = generate_neurite_locations_on(
            morphology,
            n_centers=1,
            n_per_center=self.number_of_locations,
            srcs_per_center=self.n_groups,
            center_path_distances_mean=0.0,
            center_path_distances_sd=0.0,
            max_dist_from_center=None,
            lst_section_types=self.section_types,
            seed=self.random_seed,
        ).drop(columns=[_CEN_IDX])
        return locs

    def _check_parameter_values(self):
        # Only check whenever list are resolved to individual objects
        if not isinstance(self.n_groups, list):
            assert self.n_groups > 0, "Number of groups must be at least one!"


class PathDistanceMorphologyLocations(MorphologyLocationsBlock):
    """Locations around a specified path distance."""

    path_dist_mean: float | list[float] = Field(
        name="Path distance mean",
        description="Mean of a Gaussian, defined on soma path distance in um. Used to determine \
            locations.",
    )
    path_dist_tolerance: float | list[float] = Field(
        name="Path distance tolerance",
        description="Amount of deviation in um from mean path distance that is tolerated. Must be \
            > 1.0",
    )

    def _make_points(self, morphology: morphio.Morphology):
        locs = generate_neurite_locations_on(
            morphology,
            n_centers=self.number_of_locations,
            n_per_center=1,
            srcs_per_center=1,
            center_path_distances_mean=self.path_dist_mean,
            center_path_distances_sd=0.1 * self.path_dist_tolerance,
            max_dist_from_center=0.9 * self.path_dist_tolerance,
            lst_section_types=self.section_types,
            seed=self.random_seed,
        ).drop(columns=[_CEN_IDX])
        return locs

    def _check_parameter_values(self):
        # Only check whenever list are resolved to individual objects
        if not isinstance(self.path_dist_mean, list):
            assert self.path_dist_mean >= 0.0, "Path distance mean must be non-negative!"

        if not isinstance(self.path_dist_tolerance, list):
            assert self.path_dist_tolerance >= 1.0, (
                "For numerical reasons, path distance tolerance must be at least 1.0!"
            )


class ClusteredMorphologyLocations(MorphologyLocationsBlock):
    """Clustered random locations."""

    n_clusters: int | list[int] = Field(
        name="Number of clusters", description="Number of location clusters to generate"
    )
    cluster_max_distance: float | list[float] = Field(
        name="Cluster maximum distance",
        description="Maximum distance in um of generated locations from the center of their \
            cluster",
    )

    def _make_points(self, morphology: morphio.Morphology):
        # FIXME: This rounds down. Could make missing points
        # in a second call to generate_neurite_locations_on
        n_per_cluster = int(self.number_of_locations / self.n_clusters)
        locs = generate_neurite_locations_on(
            morphology,
            n_centers=self.n_clusters,
            n_per_center=n_per_cluster,
            srcs_per_center=1,
            center_path_distances_mean=0.0,
            center_path_distances_sd=1e20,
            max_dist_from_center=self.cluster_max_distance,
            lst_section_types=self.section_types,
            seed=self.random_seed,
        ).drop(columns=[_CEN_IDX])
        return locs

    def _check_parameter_values(self):
        # Only check whenever list are resolved to individual objects
        if not isinstance(self.n_clusters, list):
            assert self.n_clusters >= 1, "Number of clusters must be at least one!"
            if not isinstance(self.number_of_locations, list):
                assert self.number_of_locations >= self.n_clusters, (
                    "Cannot make more clusters than locations!"
                )


class ClusteredGroupedMorphologyLocations(
    ClusteredMorphologyLocations, RandomGroupedMorphologyLocations
):
    """Clustered random locations, grouped in to conceptual groups."""

    def _make_points(self, morphology: morphio.Morphology):
        # FIXME: This rounds down. Could make missing points
        # in a second call to generate_neurite_locations_on
        n_per_cluster = int(self.number_of_locations / self.n_clusters)
        locs = generate_neurite_locations_on(
            morphology,
            n_centers=self.n_clusters,
            n_per_center=n_per_cluster,
            srcs_per_center=self.n_groups,
            center_path_distances_mean=0.0,
            center_path_distances_sd=1e20,
            max_dist_from_center=self.cluster_max_distance,
            lst_section_types=self.section_types,
            seed=self.random_seed,
        ).drop(columns=[_CEN_IDX])
        return locs

    def _check_parameter_values(self):
        super(ClusteredMorphologyLocations, self)._check_parameter_values()
        super(RandomGroupedMorphologyLocations, self)._check_parameter_values()


class ClusteredPathDistanceMorphologyLocations(ClusteredMorphologyLocations):
    """Clustered random locations around a specified path distance. Also creates
    groups within each cluster. This exposes the full possible complexity.
    """

    path_dist_mean: float | list[float] = Field(
        name="Path distance mean",
        description="Mean of a Gaussian, defined on soma path distance in um. Used to determine \
            locations.",
    )
    path_dist_sd: float | list[float] = Field(
        name="Path distance mean",
        description="SD of a Gaussian, defined on soma path distance in um. Used to determine \
            locations.",
    )
    n_groups_per_cluster: int | list[int] = Field(
        default=1,
        name="Number of groups per cluster",
        description="Number of conceptual groups per location cluster to generate",
    )

    def _make_points(self, morphology: morphio.Morphology):
        # FIXME: This rounds down. Could make missing points
        # in a second call to generate_neurite_locations_on
        n_per_cluster = int(self.number_of_locations / self.n_clusters)
        locs = generate_neurite_locations_on(
            morphology,
            n_centers=self.n_clusters,
            n_per_center=n_per_cluster,
            srcs_per_center=self.n_groups_per_cluster,
            center_path_distances_mean=self.path_dist_mean,
            center_path_distances_sd=self.path_dist_sd,
            max_dist_from_center=self.cluster_max_distance,
            lst_section_types=self.section_types,
            seed=self.random_seed,
        )
        return locs

    def _check_parameter_values(self):
        super()._check_parameter_values()
        # Only check whenever list are resolved to individual objects
        if not isinstance(self.path_dist_mean, list):
            assert self.path_dist_mean >= 0.0, "Path distance mean must be non-negative!"

        if not isinstance(self.path_dist_sd, list):
            assert self.path_dist_sd >= 0.1, (
                "For numerical reasons, path distance standard deviation must be at least 0.1!"
            )

        if not isinstance(self.n_groups_per_cluster, list):
            assert self.n_groups_per_cluster >= 1, (
                "Number of groups per cluster must be at least 1!"
            )



import abc
import json
import os
from typing import Annotated, Self, Literal, Optional, ClassVar

import bluepysnap as snap
import numpy as np
import pandas
import logging
from pydantic import Field, NonNegativeFloat, NonNegativeInt, model_validator, field_validator

from obi_one.core.base import OBIBaseModel
from obi_one.core.block import Block
from obi_one.core.tuple import NamedTuple
from obi_one.scientific.circuit.circuit import Circuit

from obi_one.core.exception import OBIONE_Error

L = logging.getLogger("obi-one")
_NBS1_VPM_NODE_POP = "VPM"
_NBS1_POM_NODE_POP = "POm"
_RCA1_CA3_NODE_POP = "CA3_projections"

_ALL_NODE_SET = "All"
_EXCITATORY_NODE_SET = "Excitatory"
_INHIBITORY_NODE_SET = "Inhibitory"


class NeuronPropertyFilter(OBIBaseModel, abc.ABC):
    filter_dict: dict[str, list] = Field(
        name="Filter",
        description="Filter dictionary. Note as this is NOT a Block and the list here is \
                    not to support multi-dimensional parameters but to support a key-value pair \
                    with multiple values i.e. {'layer': ['2', '3']}}",
        default={},
    )

    @model_validator(mode="after")
    def check_filter_dict_values(self) -> Self:
        for key, values in self.filter_dict.items():
            assert isinstance(values, list) and len(values) >= 1, (
                f"Filter key '{key}' must have a non-empty list of values."
            )
        return self

    @property
    def filter_keys(self) -> list[str]:
        return list(self.filter_dict.keys())

    @property
    def filter_values(self) -> list[list]:
        return list(self.filter_dict.values())

    def filter(self, df_in, reindex=True) -> pandas.DataFrame:
        ret = df_in
        for filter_key, filter_value in self.filter_dict.items():
            filter_value = [str(_entry) for _entry in filter_value]
            vld = ret[filter_key].astype(str).isin(filter_value)
            ret = ret.loc[vld]
            if reindex:
                ret = ret.reset_index(drop=True)
        return ret

    def test_validity(self, circuit, node_population: str) -> None:
        circuit_prop_names = circuit.sonata_circuit.nodes[node_population].property_names
        # filter_keys = list(self.filter_dict.keys())

        assert all(_prop in circuit_prop_names for _prop in self.filter_keys), (
            f"Invalid neuron properties! Available properties: {circuit_prop_names}"
        )

    def __repr__(self) -> str:
        """Return a string representation of the NeuronPropertyFilter object."""
        if len(self.filter_dict) == 0:
            return "NoFilter"
        string_rep = ""
        for filter_key, filter_value in self.filter_dict.items():
            string_rep += f"{filter_key}="
            for value in filter_value:
                string_rep += f"{value},"
        return string_rep[:-1]  # Remove trailing comma and space


class AbstractNeuronSet(Block, abc.ABC):
    """Base class representing a neuron set which can be turned into a SONATA node set by either
    adding it to an existing SONATA circuit object (add_node_set_to_circuit) or writing it to a
    SONATA node set .json file (write_circuit_node_set_file).
    Whenever such a neuron set is used in a SimulationsForm, it must be added to its neuron_sets
    dictionary with the key being the name of the SONATA node set which will internally be set
    in simulation_level_name upon initialization of the SimulationsForm.
    """

    sample_percentage: Annotated[NonNegativeFloat, Field(le=100)] | Annotated[list[Annotated[NonNegativeFloat, Field(le=100)]], Field(min_length=1)] = Field(
        default=100.0,
        title="Sample (Percentage)",
        description="Percentage of neurons to sample between 0 and 100%",
        units='%'
    )

    sample_seed: int | list[int] = Field(
        default=1,
        title="Sample Seed",
        description="Seed for random sampling."
    )

    @abc.abstractmethod
    def _get_expression(self, circuit: Circuit, population: str) -> dict:
        """Returns the SONATA node set expression (w/o subsampling)."""

    @staticmethod
    def check_population(circuit: Circuit, population: str) -> None:
        assert population in Circuit.get_node_population_names(circuit.sonata_circuit), (
            f"Node population '{population}' not found in circuit '{circuit}'!"
        )

    @staticmethod
    def add_node_set_to_circuit(
        sonata_circuit: snap.Circuit, node_set_dict, overwrite_if_exists=False
    ) -> None:
        """Adds the node set definition to a SONATA circuit object to make it accessible \
            (in-place).
        """
        existing_node_sets = sonata_circuit.node_sets.content
        if not overwrite_if_exists:
            for _k in node_set_dict.keys():
                assert _k not in existing_node_sets, f"Node set '{_k}' already exists!"
        existing_node_sets.update(node_set_dict)
        sonata_circuit.node_sets = snap.circuit.NodeSets.from_dict(existing_node_sets)

    @staticmethod
    def write_circuit_node_set_file(
        sonata_circuit: snap.Circuit, output_path, file_name=None, overwrite_if_exists=False
    ) -> None:
        """Writes a new node set file of a given SONATA circuit object."""
        if file_name is None:
            # Use circuit's node set file name by default
            file_name = os.path.split(sonata_circuit.config["node_sets_file"])[1]
        else:
            assert isinstance(file_name, str) and len(file_name) > 0, (
                "File name must be a non-empty string! Can be omitted to use default file name."
            )
            fname, fext = os.path.splitext(file_name)
            assert len(fname) > 0 and fext.lower() == ".json", (
                "File name must be non-empty and of type .json!"
            )
        output_file = os.path.join(output_path, file_name)

        if not overwrite_if_exists:
            assert not os.path.exists(output_file), (
                f"Output file '{output_file}' already exists! Delete file or choose to overwrite."
            )

        with open(output_file, "w") as f:
            json.dump(sonata_circuit.node_sets.content, f, indent=2)
    
    def _population(self, population: str | None=None):
        assert population is not None, "Must specify a node population name!"
        return population

    def _resolve_ids(self, circuit: Circuit, population: str | None=None) -> list[int]:
        """Returns the full list of neuron IDs (w/o subsampling)."""
        population = self._population(population)
        c = circuit.sonata_circuit
        expression = self._get_expression(circuit, population)
        name = "__TMP_NODE_SET__"
        self.add_node_set_to_circuit(c, {name: expression})
        return c.nodes[population].ids(name)

    def get_neuron_ids(self, circuit: Circuit, population: str | None=None):
        """Returns list of neuron IDs (with subsampling, if specified)."""
        self.enforce_no_lists()
        population = self._population(population)
        self.check_population(circuit, population)
        ids = np.array(self._resolve_ids(circuit, population))
        if len(ids) > 0 and self.sample_percentage < 100.0:
            rng = np.random.default_rng(self.sample_seed)

            num_sample = np.round((self.sample_percentage/100.0) * len(ids)).astype(int)

            ids = ids[
                rng.permutation([True] * num_sample + [False] * (len(ids) - num_sample))
            ]

        return ids

    def get_node_set_definition(
        self, circuit: Circuit, population: str | None=None, force_resolve_ids=False
    ) -> dict:
        """Returns the SONATA node set definition, optionally forcing to resolve individual \
            IDs.
        """
        self.enforce_no_lists()
        population = self._population(population)
        self.check_population(circuit, population)
        if self.sample_percentage is None and not force_resolve_ids:
            # Symbolic expression can be preserved
            expression = self._get_expression(circuit, population)
        else:
            # Individual IDs need to be resolved
            expression = {
                "population": population,
                "node_id": self.get_neuron_ids(circuit, population).tolist(),
            }

        return expression
    
    def population_type(self, circuit: Circuit, population: str | None=None):
        """Returns the population type (i.e. biophysical / virtual)."""
        return circuit.sonata_circuit.nodes[self._population(population)].type

    def to_node_set_file(
        self,
        circuit: Circuit,
        population,
        output_path,
        file_name=None,
        overwrite_if_exists=False,
        append_if_exists=False,
        force_resolve_ids=False,
        init_empty=False,
        optional_node_set_name=None,
    ):
        """Resolves the node set for a given circuit/population and writes it to a .json node \
            set file.
        """
        if optional_node_set_name is not None:
            self.set_simulation_level_name(optional_node_set_name)
        assert self.name is not None, "NeuronSet name must be set through the Simulation or optional_node_set_name parameter!"

        if file_name is None:
            # Use circuit's node set file name by default
            file_name = os.path.split(circuit.sonata_circuit.config["node_sets_file"])[1]
        else:
            assert isinstance(file_name, str) and len(file_name) > 0, (
                "File name must be a non-empty string! Can be omitted to use default file name."
            )
            fname, fext = os.path.splitext(file_name)
            assert len(fname) > 0 and fext.lower() == ".json", (
                "File name must be non-empty and of type .json!"
            )
        output_file = os.path.join(output_path, file_name)

        assert not (overwrite_if_exists and append_if_exists), (
            "Append and overwrite options are mutually exclusive!"
        )
        population = self._population(population)
        expression = self.get_node_set_definition(
            circuit, population, force_resolve_ids=force_resolve_ids
        )
        assert expression is not None, "Node set already exists in circuit, nothing to be done!"

        if not os.path.exists(output_file) or overwrite_if_exists:
            # Create new node sets file, overwrite if existing
            if init_empty:
                # Initialize empty
                node_sets = {}
            else:
                # Initialize with circuit object's node sets
                node_sets = circuit.sonata_circuit.node_sets.content
                assert self.name not in node_sets, (
                    f"Node set '{self.name}' already exists in circuit '{circuit}'!"
                )
            node_sets.update({self.name: expression})

        elif os.path.exists(output_file) and append_if_exists:
            # Append to existing node sets file
            with open(output_file) as f:
                node_sets = json.load(f)
                assert self.name not in node_sets, (
                    f"Appending not possible, node set '{self.name}' already exists!"
                )
                node_sets.update({self.name: expression})

        else:  # File existing but no option chosen
            assert False, (
                f"Output file '{output_file}' already exists! Delete file or choose to append \
                    or overwrite."
            )

        with open(output_file, "w") as f:
            json.dump(node_sets, f, indent=2)

        return output_file


class NeuronSet(AbstractNeuronSet):
    """
    Extension of abstract neuron set with the ability to specify the node population upon creation.
    This is optional, all functions requiring a node population can be optionally called with the name of
    a default population to be used in case no name was set upon creation.
    """
    node_population: str | list[str] | None = None

    def _population(self, population: str | None=None):
        if population is not None and self.node_population is not None:
            if population != self.node_population:
                L.warning("Node population %s has been set for this block and will be used. Ignoring %s",
                          self.node_population, population)
        population = self.node_population or population
        if population is None:
            raise ValueError("Must specify name of a node population to resolve the NeuronSet!")
        return population
    

class PredefinedNeuronSet(NeuronSet):
    """Neuron set wrapper of an existing (named) node sets already predefined in the node \
        sets file.
    """

    node_set: (
        Annotated[str, Field(min_length=1)]
        | Annotated[list[Annotated[str, Field(min_length=1)]], Field(min_length=1)]
    )

    def check_node_set(self, circuit: Circuit, population: str) -> None:
        assert self.node_set in circuit.node_sets, (
            f"Node set '{self.node_set}' not found in circuit '{circuit}'!"
        )  # Assumed that all (outer) lists have been resolved

    def _get_expression(self, circuit: Circuit, population):
        """Returns the SONATA node set expression (w/o subsampling)."""
        self.check_node_set(circuit, population)
        return [self.node_set]


class AllNeurons(AbstractNeuronSet):
    """All biophysical neurons."""

    title: ClassVar[str] = "All Neurons"
    
    def check_node_set(self, circuit: Circuit, population: str) -> None:
        assert _ALL_NODE_SET in circuit.node_sets, (
            f"Node set '{_ALL_NODE_SET}' not found in circuit '{circuit}'!"
        )  # Assumed that all (outer) lists have been resolved

    def _get_expression(self, circuit: Circuit, population):
        """Returns the SONATA node set expression (w/o subsampling)."""
        self.check_node_set(circuit, population)
        return [_ALL_NODE_SET]


class ExcitatoryNeurons(AbstractNeuronSet):
    """All biophysical excitatory neurons."""

    title: ClassVar[str] = "All Excitatory Neurons"

    def check_node_set(self, circuit: Circuit, population: str) -> None:
        assert _EXCITATORY_NODE_SET in circuit.node_sets, (
            f"Node set '{_EXCITATORY_NODE_SET}' not found in circuit '{circuit}'!"
        )  # Assumed that all (outer) lists have been resolved

    def _get_expression(self, circuit: Circuit, population):
        """Returns the SONATA node set expression (w/o subsampling)."""
        self.check_node_set(circuit, population)
        return [_EXCITATORY_NODE_SET]


class InhibitoryNeurons(AbstractNeuronSet):
    """All inhibitory neurons."""

    title: ClassVar[str] = "All Inhibitory Neurons"

    def check_node_set(self, circuit: Circuit, population: str) -> None:
        assert _INHIBITORY_NODE_SET in circuit.node_sets, (
            f"Node set '{_INHIBITORY_NODE_SET}' not found in circuit '{circuit}'!"
        )  # Assumed that all (outer) lists have been resolved

    def _get_expression(self, circuit: Circuit, population):
        """Returns the SONATA node set expression (w/o subsampling)."""
        self.check_node_set(circuit, population)
        return [_INHIBITORY_NODE_SET]


class nbS1VPMInputs(AbstractNeuronSet):
    """Virtual neurons projecting from the VPM thalamic nucleus to biophysical cortical neurons in the nbS1 model."""

    title: ClassVar[str] = "Demo: nbS1 VPM Inputs"
    
    def _population(self, population: str | None=None):
        # Ignore default node population name. This is always VPM.
        return _NBS1_VPM_NODE_POP
    
    def _get_expression(self, circuit: Circuit, population):
        return {"population": _NBS1_VPM_NODE_POP}
    

class nbS1POmInputs(AbstractNeuronSet):
    """Virtual neurons projecting from the POm thalamic nucleus to biophysical cortical neurons in the nbS1 model."""

    title: ClassVar[str] = "Demo: nbS1 POm Inputs"
    
    def _population(self, population: str | None=None):
        # Ignore default node population name. This is always POm.
        return _NBS1_POM_NODE_POP
    
    def _get_expression(self, circuit: Circuit, population):
        return {"population": _NBS1_POM_NODE_POP}


class rCA1CA3Inputs(AbstractNeuronSet):
    """Virtual neurons projecting from the CA3 region to biophysical CA1 neurons in the rCA1 model."""

    title: ClassVar[str] = "Demo: rCA1 CA3 Inputs"
    
    def _population(self, population: str | None=None):
        # Ignore default node population name. This is always CA3_projections.
        return _RCA1_CA3_NODE_POP
    
    def _get_expression(self, circuit: Circuit, population):
        return {"population": _RCA1_CA3_NODE_POP}
    

class CombinedNeuronSet(NeuronSet):
    """Neuron set definition based on a combination of existing (named) node sets."""

    node_sets: (
        Annotated[tuple[Annotated[str, Field(min_length=1)], ...], Field(min_length=1)]
        | Annotated[
            list[Annotated[tuple[Annotated[str, Field(min_length=1)], ...], Field(min_length=1)]],
            Field(min_length=1),
        ]
    )

    def check_node_sets(self, circuit: Circuit, population: str) -> None:
        for _nset in self.node_sets:  # Assumed that all (outer) lists have been resolved
            assert _nset in circuit.node_sets, (
                f"Node set '{_nset}' not found in circuit '{circuit}'!"
            )

    def _get_expression(self, circuit: Circuit, population: str):
        """Returns the SONATA node set expression (w/o subsampling)."""
        self.check_node_sets(circuit, population)
        return list(self.node_sets)


class IDNeuronSet(AbstractNeuronSet):
    """Neuron set definition by providing a list of neuron IDs."""

    title: ClassVar[str] = "ID Neuron Set"

    neuron_ids: NamedTuple | Annotated[list[NamedTuple], Field(min_length=1)]

    def check_neuron_ids(self, circuit: Circuit, population) -> None:
        popul_ids = circuit.sonata_circuit.nodes[population].ids()
        assert all(_nid in popul_ids for _nid in self.neuron_ids.elements), (
            f"Neuron ID(s) not found in population '{population}' of circuit '{circuit}'!"
        )  # Assumed that all (outer) lists have been resolved

    def _get_expression(self, circuit: Circuit, population) -> dict:
        """Returns the SONATA node set expression (w/o subsampling)."""
        population = self._population(population)
        self.check_neuron_ids(circuit, population)
        return {"population": population, "node_id": list(self.neuron_ids.elements)}


class PropertyNeuronSet(NeuronSet):
    """Neuron set definition based on neuron properties, optionally combined with (named) node \
        sets.
    """

    property_filter: NeuronPropertyFilter | list[NeuronPropertyFilter] = Field(
        name="Neuron property filter",
        description="NeuronPropertyFilter object or list of NeuronPropertyFilter objects",
        default=(),
    )
    node_sets: (
        tuple[Annotated[str, Field(min_length=1)], ...]
        | Annotated[list[tuple[Annotated[str, Field(min_length=1)], ...]], Field(min_length=1)]
    ) = tuple()

    def check_properties(self, circuit: Circuit, population: str | None=None) -> None:
        population = self._population(population)
        self.property_filter.test_validity(circuit, population)

    def check_node_sets(self, circuit: Circuit, population: str) -> None:
        for _nset in self.node_sets:  # Assumed that all (outer) lists have been resolved
            assert _nset in circuit.node_sets, (
                f"Node set '{_nset}' not found in circuit '{circuit}'!"
            )

    def _get_resolved_expression(self, circuit: Circuit, population: str | None=None) -> dict:
        """A helper function used to make subclasses work."""
        c = circuit.sonata_circuit
        population = self._population(population)

        df = c.nodes[population].get(properties=self.property_filter.filter_keys).reset_index()
        df = self.property_filter.filter(df)

        node_ids = df["node_ids"].values

        if len(self.node_sets) > 0:
            node_ids_nset = np.array([]).astype(int)
            for _nset in self.node_sets:
                node_ids_nset = np.union1d(node_ids_nset, c.nodes[population].ids(_nset))
            node_ids = np.intersect1d(node_ids, node_ids_nset)

        expression = {"population": population, "node_id": node_ids.tolist()}
        return expression

    def _get_expression(self, circuit: Circuit, population) -> dict:
        """Returns the SONATA node set expression (w/o subsampling)."""
        population = self._population(population)
        self.check_properties(circuit, population)
        self.check_node_sets(circuit, population)

        def __resolve_sngl(prop_vals):
            if len(prop_vals) == 1:
                return prop_vals[0]
            return list(prop_vals)

        if len(self.node_sets) == 0:
            # Symbolic expression can be preserved
            expression = dict(
                [
                    (property_key, __resolve_sngl(property_value))
                    for property_key, property_value in self.property_filter.filter_dict.items()
                ]
            )
        else:
            # Individual IDs need to be resolved
            return self._get_resolved_expression(circuit, population)

        return expression


class VolumetricCountNeuronSet(PropertyNeuronSet):
    """Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat."""

    ox: float | list[float] = Field(
        name="Offset: x",
        description="Offset of the center of the volume, relative to the centroid of the node \
            population",
    )
    oy: float | list[float] = Field(
        name="Offset: y",
        description="Offset of the center of the volume, relative to the centroid of the node \
            population",
    )
    oz: float | list[float] = Field(
        name="Offset: z",
        description="Offset of the center of the volume, relative to the centroid of the node \
            population",
    )
    n: NonNegativeInt | list[NonNegativeInt] = Field(
        name="Number of neurons", description="Number of neurons to find"
    )
    columns_xyz: tuple[str, str, str] | list[tuple[str, str, str]] = Field(
        name="x/y/z column names",
        description="Names of the three neuron (node) properties used for volumetric tests",
        default=("x", "y", "z"),
    )

    def _get_expression(self, circuit: Circuit, population: str | None=None) -> dict:
        population = self._population(population)
        self.check_properties(circuit, population)
        self.check_node_sets(circuit, population)
        # Always needs to be resolved
        base_expression = self._get_resolved_expression(circuit, population)

        cols_xyz = list(self.columns_xyz)
        df = circuit.sonata_circuit.nodes[population].get(
            base_expression["node_id"], properties=cols_xyz
        )
        df = df.reset_index(drop=False)
        o_df = pandas.Series({cols_xyz[0]: self.ox, cols_xyz[1]: self.oy, cols_xyz[2]: self.oz})
        tgt_center = df[cols_xyz].mean() + o_df

        D = np.linalg.norm(df[cols_xyz] - tgt_center, axis=1)
        idxx = np.argsort(D)[: self.n]
        df = df.iloc[idxx]

        expression = {"population": population, "node_id": list(df["node_ids"].astype(int))}
        return expression

class SimplexMembershipBasedNeuronSet(PropertyNeuronSet):
    """
    Sample neurons from the set of neurons that form simplices of a given dimension with a chosen source or target 'central' neuron.
    """
    central_neuron_id: int = Field(
        name="Central neuron id",
        description="Node id (index) that will be source or target of the simplices extracted",
    )
    dim: int = Field(
        name="Dimension",
        description="Dimension of the simplices to be extracted",
    )
    central_neuron_simplex_position: Literal['source', 'target'] = Field('source',
        name="Central neuron simplex position",
        description="Position of the central neuron/node in the simplex, it can be either 'source' or 'target'",
    )
    subsample: bool = Field(True,
        name="subsample",
        description="Whether to subsample the set of nodes in the simplex lists or not",
    )
    n_count_max: Optional[int] = Field(False,
        name="Max node count",
        description="Maximum number of nodes to be subsampled",
    )
    subsample_method: Literal['node_participation', 'random'] = Field('node_participation',
        name="Method to subsample nodes from the extracted simplices",
        description="""
        **Method to subsample nodes**:
        - `random`: randomly selects nodes from all nodes in the simplices
        - `node_participation`: selects nodes with highest node participation 
            """,
    )
    simplex_type: Literal['directed', 'reciprocal', 'undirected'] = Field('directed',
        name="Simplex type",
        description="Type of simplex to consider. See more at \
            https://openbraininstitute.github.io/connectome-analysis/network_topology/#src.connalysis.network.topology.simplex_counts",
    )
    seed: Optional[int] = Field(None,
        name="seed",
        description="Seed used for random subsampling method",
    )

    @field_validator('dim')
    def dim_check(cls, v):
        if v <= 1:
            raise ValueError('Simplex dimension must be greater than 1')
        return v
    @model_validator(mode="after")
    def check_n_count_max(self) -> Self:
        n_count_max = self.n_count_max
        if self.subsample and n_count_max is None:
            raise ValueError("n_count_max must be specified when subsample is True")
        return self

    def _get_expression(self, circuit: Circuit, population: str) -> dict:
        try: # Try to import connalysis
            from obi_one.scientific.circuit.simplex_extractors import (simplex_submat)
        except ImportError as e:
            print(f"Import failed: {e}. You probably need to install connalysis locally")

        self.check_properties(circuit, population)
        self.check_node_sets(circuit, population)
        # Always needs to be resolved
        base_expression = self._get_resolved_expression(circuit, population)

        # Restrict connectivity matrix to chosen subpopulation and find index of center
        conn=circuit.connectivity_matrix.subpopulation(base_expression["node_id"])
        index = np.where(conn.vertices["node_ids"] == self.central_neuron_id)[0]
        if len(index) == 0:
            raise ValueError(f"The neuron of index {self.central_neuron_id} is not in the subpopulation selected")
        index = index[0]    
            
        # Get nodes on simplices index by 0 ... conn._shape[0] 
        out = simplex_submat(
            conn.matrix, index, self.dim, self.central_neuron_simplex_position,
            self.subsample, self.n_count_max, self.subsample_method,
            self.simplex_type, self.seed
        )
        selection = out[0] if self.subsample else out

        # Get node_ids (i.e., get correct index) and build expression dict 
        selection=conn.vertices["node_ids"].iloc[selection]
        expression = {"population": population, "node_id": selection.tolist()}        
        return expression
    

class SimplexNeuronSet(PropertyNeuronSet):
    """
    Get neurons that form simplices of a given dimension with a chosen source or target neuron.
    If a smaller sample is required, it samples simplices while the number of nodes on them is still 
    smaller or equal than a set target size.
    """
    central_neuron_id: int = Field(
        name="Central neuron id",
        description="Node id (index) that will be source or target of the simplices extracted",
    )
    dim: int = Field(
        name="Dimension",
        description="Dimension of the simplices to be extracted",
    )
    central_neuron_simplex_position: Literal['source', 'target'] = Field('source',
        name="Central neuron simplex position",
        description="Position of the central neuron/node in the simplex, it can be either 'source' or 'target'",
    )
    subsample: bool = Field(False,
        name="subsample",
        description="Whether to subsample the set of nodes in the simplex lists or not",
    )
    n_count_max: Optional[int] = Field(None,
        name="Max node count",
        description="Maximum number of nodes to be subsampled",
    )
    simplex_type: Literal['directed', 'reciprocal', 'undirected'] = Field('directed',
        name="Simplex type",
        description="Type of simplex to consider. See more at \
            https://openbraininstitute.github.io/connectome-analysis/network_topology/#src.connalysis.network.topology.simplex_counts",
    )
    seed: Optional[int] = Field(None,
        name="seed",
        description="Seed used for random subsampling method",
    )

    @field_validator('dim')
    def dim_check(cls, v):
        if v <= 1:
            raise ValueError('Simplex dimension must be greater than 1')
        return v
    @model_validator(mode="after")
    def check_n_count_max(self) -> Self:
        n_count_max = self.n_count_max
        if self.subsample and n_count_max is None:
            raise ValueError("n_count_max must be specified when subsample is True")
        return self

    def _get_expression(self, circuit: Circuit, population: str) -> dict:
        try: # Try to import connalysis
            from obi_one.scientific.circuit.simplex_extractors import (simplex_submat)
        except ImportError as e:
            print(f"Import failed: {e}. You probably need to install connalysis locally")

        self.check_properties(circuit, population)
        self.check_node_sets(circuit, population)
        # Always needs to be resolved
        base_expression = self._get_resolved_expression(circuit, population)

        # Restrict connectivity matrix to chosen subpopulation and find index of center
        conn=circuit.connectivity_matrix.subpopulation(base_expression["node_id"])
        index = np.where(conn.vertices["node_ids"] == self.central_neuron_id)[0]
        if len(index) == 0:
            raise ValueError(f"The neuron of index {self.central_neuron_id} is not in the subpopulation selected")
        index = index[0]    
            
        # Get nodes on simplices index by 0 ... conn._shape[0] 
        out = simplex_submat(
            conn.matrix, index, self.dim, self.central_neuron_simplex_position,
            self.subsample, self.n_count_max, 'sample_simplices',
            self.simplex_type, self.seed
        )
        selection = out[0] if self.subsample else out

        # Get node_ids (i.e., get correct index) and build expression dict 
        selection=conn.vertices["node_ids"].iloc[selection]
        expression = {"population": population, "node_id": selection.tolist()}        
        return expression
    



class VolumetricRadiusNeuronSet(PropertyNeuronSet):
    """Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat."""

    ox: float | list[float] = Field(
        name="Offset: x",
        description="Offset of the center of the volume, relative to the centroid of the node \
            population",
    )
    oy: float | list[float] = Field(
        name="Offset: y",
        description="Offset of the center of the volume, relative to the centroid of the node \
            population",
    )
    oz: float | list[float] = Field(
        name="Offset: z",
        description="Offset of the center of the volume, relative to the centroid of the node \
            population",
    )
    radius: NonNegativeFloat | list[NonNegativeFloat] = Field(
        name="Radius", description="Radius in um of volumetric sample"
    )
    columns_xyz: tuple[str, str, str] | list[tuple[str, str, str]] = Field(
        name="x/y/z column names",
        description="Names of the three neuron (node) properties used for volumetric tests",
        default=("x", "y", "z"),
    )

    def _get_expression(self, circuit: Circuit, population: str | None=None) -> dict:
        population = self._population(population)
        self.check_node_sets(circuit, population)
        # Always needs to be resolved
        base_expression = self._get_resolved_expression(circuit, population)

        cols_xyz = list(self.columns_xyz)
        df = circuit.sonata_circuit.nodes[population].get(
            base_expression["node_id"], properties=cols_xyz
        )
        df = df.reset_index(drop=False)
        o_df = pandas.Series({cols_xyz[0]: self.ox, cols_xyz[1]: self.oy, cols_xyz[2]: self.oz})
        tgt_center = df[cols_xyz].mean() + o_df

        D = np.linalg.norm(df[cols_xyz] - tgt_center, axis=1)
        idxx = np.nonzero(self.radius > D)[0]
        df = df.iloc[idxx]

        expression = {"population": population, "node_id": list(df["node_ids"].astype(int))}
        return expression



import json
import logging
from collections import OrderedDict
from importlib.metadata import version
from pathlib import Path
from typing import Any

from pydantic import Field, field_validator

from obi_one.core.base import OBIBaseModel
from obi_one.core.block import Block
from obi_one.core.param import SingleValueScanParam

L = logging.getLogger(__name__)


class SingleCoordinateScanParams(OBIBaseModel):
    scan_params: list[SingleValueScanParam] = Field(default_factory=list)
    nested_coordinate_subpath_str: Path = Path()

    @property
    def nested_param_name_and_value_subpath(self) -> Path:
        if len(self.scan_params):
            self.nested_coordinate_subpath_str = ""
            for scan_param in self.scan_params:
                self.nested_coordinate_subpath_str += (
                    f"{scan_param.location_str}={scan_param.value}/"
                )
            return Path(self.nested_coordinate_subpath_str)
        return Path(self.nested_coordinate_subpath_str)

    @property
    def nested_param_value_subpath(self) -> Path:
        if len(self.scan_params):
            self.nested_coordinate_subpath_str = ""
            for scan_param in self.scan_params:
                self.nested_coordinate_subpath_str += f"{scan_param.value}/"
            return Path(self.nested_coordinate_subpath_str)
        return Path(self.nested_coordinate_subpath_str)

    def display_parameters(self) -> None:
        output = ""

        if len(self.scan_params) == 0:
            L.info("No coordinate parameters.")
        else:
            for j, scan_param in enumerate(self.scan_params):
                output = output + scan_param.location_str + ": " + str(scan_param.value)
                if j < len(self.scan_params) - 1:
                    output += ", "
            L.info(output)

    def dictionary_representaiton(self) -> dict[str, Any]:
        """Return a dictionary representation of the scan parameters."""
        d = {}
        for scan_param in self.scan_params:
            d[scan_param.location_str] = scan_param.value
        return d


class SingleCoordinateMixin:
    """Mixin to enforce no lists in all Blocks and Blocks in Category dictionaries."""

    idx: int = -1
    scan_output_root: Path = Path()
    coordinate_output_root: Path = Path()
    _coordinate_directory_option: str = "NAME_EQUALS_VALUE"
    single_coordinate_scan_params: SingleCoordinateScanParams = None

    @field_validator("*", mode="before")
    @classmethod
    def enforce_single_type(cls, value: Any) -> Any:
        if isinstance(value, dict):  # For Block instances in 1st level dictionaries
            for dict_value in value.values():
                if isinstance(dict_value, Block):
                    block = dict_value
                    block.enforce_no_lists()  # Enforce no lists

        if isinstance(value, Block):  # For Block instances at the 1st level
            block = value
            value.enforce_no_lists()  # Enforce no lists

        return value

    def initialize_coordinate_output_root(
        self, scan_output_root: Path, coordinate_directory_option: str = "NAME_EQUALS_VALUE"
    ) -> None:
        """Initialize the output root paths for the scan and coordinate directories."""
        self.scan_output_root = scan_output_root

        self._coordinate_directory_option = coordinate_directory_option

        if self._coordinate_directory_option == "NAME_EQUALS_VALUE":
            self.coordinate_output_root = (
                self.scan_output_root
                / self.single_coordinate_scan_params.nested_param_name_and_value_subpath
            )

        elif self._coordinate_directory_option == "VALUE":
            self.coordinate_output_root = (
                self.scan_output_root
                / self.single_coordinate_scan_params.nested_param_value_subpath
            )
        elif self._coordinate_directory_option == "ZERO_INDEX":
            self.coordinate_output_root = self.scan_output_root / f"{self.idx}"
        else:
            msg = (
                f"Invalid coordinate_directory_option: {self._coordinate_directory_option}. "
                "Valid options are: NAME_EQUALS_VALUE, VALUE, ZERO_INDEX."
            )
            raise ValueError(msg)

        # Create the coordinate_output_root directory
        Path.mkdir(self.coordinate_output_root, parents=True, exist_ok=True)

    def serialize(self, output_path: Path) -> None:
        """Serialize the object to a JSON file."""
        # Important to use model_dump_json() instead of model_dump()
        # so OBIBaseModel's custom encoder is used to seri
        # PosixPaths as strings
        model_dump = self.model_dump_json()

        # Now load it back into a dict to do some additional modifications
        model_dump = OrderedDict(json.loads(model_dump))

        model_dump["obi_one_version"] = version("obi-one")
        model_dump.move_to_end("scan_output_root", last=False)
        model_dump.move_to_end("coordinate_output_root", last=False)
        model_dump.move_to_end("idx", last=False)
        model_dump.move_to_end("type", last=False)
        model_dump.move_to_end("obi_one_version", last=False)

        with Path.open(output_path, "w") as json_file:
            json.dump(model_dump, json_file, indent=4)



from typing import ClassVar, Literal

from obi_one.core.block import Block
from obi_one.core.form import Form
from obi_one.core.path import NamedPath
from obi_one.core.single import SingleCoordinateMixin


class MorphologyDecontainerizationsForm(Form):
    """ """

    single_coord_class_name: ClassVar[str] = "MorphologyDecontainerization"
    name: ClassVar[str] = "Morphology Decontainerization"
    description: ClassVar[str] = (
        "Creates a circuit with individual morphology files instead of containerized morphologies"
    )

    class Initialize(Block):
        circuit_path: NamedPath | list[NamedPath]
        output_format: Literal["h5", "asc", "swc"] | list[Literal["h5", "asc", "swc"]] = "h5"

    initialize: Initialize


import json
import logging
import os
import shutil
import traceback
from typing import ClassVar

import h5py
import numpy as np
import tqdm
from bluepysnap import Circuit
from morph_tool import convert


class MorphologyDecontainerization(MorphologyDecontainerizationsForm, SingleCoordinateMixin):
    """Creates a circuit with individual morphology files instead of containerized morphologies,
    which involves the following steps:
    (1) Copy circuit to output location
    (2) Extract individual .h5 morphologies from an .h5 container
    (3) Convert .h5 morphologies to specified output format (.swc or .asc; skip if .h5)
    (4) Update the circuit config, pointing to the individual morphology folder
    (5) Delete .h5 container and .h5 files if that's not the specified output format
    (7) Check loading individual morphologies
    Important: The original circuit won't be modified! The circuit will be copied
               to the output location where all operations take place.
    """

    @staticmethod
    def _check_morphologies(circuit_config, extension):
        """Check modified circuit by loading some morphologies from each node population."""
        c = Circuit(circuit_config)
        for npop in c.nodes.population_names:
            nodes = c.nodes[npop]
            if nodes.type == "biophysical":
                all_nids = nodes.ids()
                if len(all_nids) < 20:
                    nid_list = all_nids  # Check all node IDs
                    print(f"Checking all morphologies in population '{npop}'")
                else:
                    nid_list = all_nids[[0, -1]]  # Check first/last node ID only
                    print(f"Checking first/last morphologies in population '{npop}'")
                for nid in nid_list:
                    try:
                        morph = nodes.morph.get(
                            nid, transform=True, extension=extension
                        )  # Will throw an error if not accessible
                    except:
                        return False  # Error
        return True  # All successful

    def run(self) -> None:
        try:
            print(f"Running morphology decontainerization for '{self.initialize.circuit_path}'")

            # Set logging level to WARNING to prevent a lot of debug output from morph_tool.convert()
            logging.getLogger("morph_tool").setLevel(logging.WARNING)

            # Copy contents of original circuit folder to output_root
            input_path, input_config = os.path.split(self.initialize.circuit_path.path)
            output_path = self.coordinate_output_root
            circuit_config = os.path.join(output_path, input_config)
            assert not os.path.exists(circuit_config), "ERROR: Output circuit already exists!"
            print("Copying circuit to output folder...")
            shutil.copytree(input_path, output_path, dirs_exist_ok=True)
            print("...DONE")

            # Load circuit at new location
            c = Circuit(circuit_config)
            node_populations = c.nodes.population_names

            # Iterate over node populations, find all morphologies and extract them from the .h5 container
            morph_folders_to_delete = []
            morph_containers_to_delete = []
            global_morph_entry = None
            for npop in node_populations:
                nodes = c.nodes[npop]
                if nodes.type != "biophysical":
                    continue
                morph_names = np.unique(nodes.get(properties="morphology"))
                print(
                    f"> {len(morph_names)} unique morphologies in population '{npop}' ({nodes.size})"
                )

                h5_container = nodes.morph._get_morphology_base(
                    "h5"
                )  # FIXME: Should not use private function!!
                assert os.path.splitext(h5_container)[1].lower() == ".h5", (
                    "ERROR: .h5 morphology path is not a container!"
                )
                h5_folder = os.path.join(os.path.split(h5_container)[0], "h5")
                target_folder = os.path.join(
                    os.path.split(h5_container)[0], self.initialize.output_format
                )
                os.makedirs(h5_folder, exist_ok=True)
                os.makedirs(target_folder, exist_ok=True)

                # Extract from .h5 container
                with h5py.File(h5_container, "r") as f_container:
                    skip_counter = 0
                    for _m in tqdm.tqdm(
                        morph_names, desc="Extracting/converting from .h5 container"
                    ):
                        _h5_file = os.path.join(h5_folder, _m + ".h5")
                        if os.path.exists(_h5_file):
                            skip_counter += 1
                        else:
                            # Create individual .h5 morphology file
                            with h5py.File(_h5_file, "w") as f_h5:
                                # Copy all groups/datasets into root of the file
                                for _key in f_container[_m].keys():
                                    f_container.copy(f_container[f"{_m}/{_key}"], f_h5)
                            # Convert to required output format
                            if self.initialize.output_format != "h5":
                                src_file = os.path.join(h5_folder, _m + ".h5")
                                dest_file = os.path.join(
                                    target_folder, _m + f".{self.initialize.output_format}"
                                )
                                if not os.path.exists(dest_file):
                                    convert(src_file, dest_file)
                print(
                    f"Extracted/converted {len(morph_names) - skip_counter} morphologies from .h5 container ({skip_counter} already existed)"
                )
                if h5_container not in morph_containers_to_delete:
                    morph_containers_to_delete.append(h5_container)
                if self.initialize.output_format != "h5":
                    if h5_folder not in morph_folders_to_delete:
                        morph_folders_to_delete.append(h5_folder)

                # Update the circuit config so that it points to the individual morphology folder,
                # keeping the original global/local config file structure as similar as it was before
                # (but removing all other references to the original morphology folders)
                cname, cext = os.path.splitext(circuit_config)
                # shutil.copy(circuit_config, cname + "__BAK__" + cext)  # Save original config file

                with open(circuit_config) as f:
                    cfg_dict = json.load(f)

                assert "manifest" in cfg_dict and "$BASE_DIR" in cfg_dict["manifest"], (
                    "ERROR: $BASE_DIR not defined!"
                )
                assert (
                    cfg_dict["manifest"]["$BASE_DIR"] == "."
                    or cfg_dict["manifest"]["$BASE_DIR"] == "./"
                ), "ERROR: $BASE_DIR is not corcuit root directory!"
                root_path = os.path.split(circuit_config)[0]
                rel_target_folder = os.path.join(
                    "$BASE_DIR", os.path.relpath(target_folder, root_path)
                )

                # Check if there is a global entry for morphologies (initially not set)
                if global_morph_entry is None:
                    global_morph_entry = False
                    if "components" in cfg_dict:
                        if "alternate_morphologies" in cfg_dict["components"]:
                            if "h5v1" in cfg_dict["components"]["alternate_morphologies"]:
                                if (
                                    len(cfg_dict["components"]["alternate_morphologies"]["h5v1"])
                                    > 0
                                ):
                                    global_morph_entry = True

                    if global_morph_entry:  # Set morphology path globally
                        if self.initialize.output_format == "h5":
                            cfg_dict["components"]["alternate_morphologies"] = {
                                "h5v1": rel_target_folder
                            }
                            if "morphologies_dir" in cfg_dict["components"]:
                                cfg_dict["components"]["morphologies_dir"] = ""
                        elif self.initialize.output_format == "asc":
                            cfg_dict["components"]["alternate_morphologies"] = {
                                "neurolucida-asc": rel_target_folder
                            }
                            if "morphologies_dir" in cfg_dict["components"]:
                                cfg_dict["components"]["morphologies_dir"] = ""
                        else:
                            cfg_dict["components"]["morphologies_dir"] = rel_target_folder
                            if "alternate_morphologies" in cfg_dict["components"]:
                                cfg_dict["components"]["alternate_morphologies"] = {}
                if not global_morph_entry:  # Set individually per population
                    for _ndict in cfg_dict["networks"]["nodes"]:
                        if nodes.name in _ndict["populations"]:
                            _pop = _ndict["populations"][nodes.name]
                            base_path = None
                            if self.initialize.output_format == "h5":
                                _pop["alternate_morphologies"] = {"h5v1": h5_folder}
                                if "morphologies_dir" in _pop:
                                    _pop["morphologies_dir"] = ""
                            elif self.initialize.output_format == "asc":
                                _pop["alternate_morphologies"] = {
                                    "neurolucida-asc": rel_target_folder
                                }
                                if "morphologies_dir" in _pop:
                                    _pop["morphologies_dir"] = ""
                            else:
                                _pop["morphologies_dir"] = rel_target_folder
                                if "alternate_morphologies" in _pop:
                                    _pop["alternate_morphologies"] = {}
                            break
                else:
                    pass  # Skip, should be already set

                with open(circuit_config, "w") as f:
                    json.dump(cfg_dict, f, indent=2)

            # Clean up morphology folders with individual morphologies
            print(f"Cleaning morphology container(s): {morph_containers_to_delete}")
            for _file in morph_containers_to_delete:
                os.remove(_file)
            print(f"Cleaning morphology folder(s): {morph_folders_to_delete}")
            for _folder in morph_folders_to_delete:
                shutil.rmtree(_folder)

            # Reload and check morphologies in modified circuit
            assert self._check_morphologies(circuit_config, self.initialize.output_format), (
                "ERROR: Morphology check not successful!"
            )
            print("Morphology decontainerization DONE")

        except Exception as e:
            traceback.print_exception(e)



from obi_one.scientific.simulation.stimulus import (
    ConstantCurrentClampSomaticStimulus,
    HyperpolarizingCurrentClampSomaticStimulus,
    LinearCurrentClampSomaticStimulus,
    MultiPulseCurrentClampSomaticStimulus,
    NormallyDistributedCurrentClampSomaticStimulus,
    RelativeNormallyDistributedCurrentClampSomaticStimulus,
    RelativeConstantCurrentClampSomaticStimulus,
    RelativeLinearCurrentClampSomaticStimulus,
    SinusoidalCurrentClampSomaticStimulus,
    SubthresholdCurrentClampSomaticStimulus,
    PoissonSpikeStimulus,
    FullySynchronousSpikeStimulus
)

StimulusUnion = (
    ConstantCurrentClampSomaticStimulus
    | RelativeConstantCurrentClampSomaticStimulus
    | LinearCurrentClampSomaticStimulus
    | RelativeLinearCurrentClampSomaticStimulus
    | NormallyDistributedCurrentClampSomaticStimulus
    | RelativeNormallyDistributedCurrentClampSomaticStimulus
    | SinusoidalCurrentClampSomaticStimulus
    | SubthresholdCurrentClampSomaticStimulus
    | HyperpolarizingCurrentClampSomaticStimulus
    | MultiPulseCurrentClampSomaticStimulus
    | PoissonSpikeStimulus
    | FullySynchronousSpikeStimulus
)

from obi_one.core.block_reference import BlockReference
from typing import ClassVar, Any
class StimulusReference(BlockReference):
    """A reference to a StimulusUnion block."""
    
    allowed_block_types: ClassVar[Any] = StimulusUnion


from obi_one.core.block import Block


class IntracellularLocationSet(Block):
    """ """

    neuron_ids: tuple[int, ...] | list[tuple[int, ...]]


class SectionIntracellularLocationSet(IntracellularLocationSet):
    """ """

    section: str | list[str]



import logging
from typing import ClassVar

L = logging.getLogger(__name__)

from obi_one.core.block import Block
from obi_one.core.form import Form
from obi_one.core.path import NamedPath
from obi_one.core.single import SingleCoordinateMixin


class FolderCompressions(Form):
    """ """

    single_coord_class_name: ClassVar[str] = "FolderCompression"
    name: ClassVar[str] = "Folder Compression"
    description: ClassVar[str] = "Compresses a folder using the specified compression format."

    class Initialize(Block):
        folder_path: NamedPath | list[NamedPath]
        file_format: None | str | list[None | str] = "gz"
        file_name: None | str | list[None | str] = "compressed"

    initialize: Initialize


import os
import tarfile
import time
import traceback
from typing import ClassVar


class FolderCompression(FolderCompressions, SingleCoordinateMixin):
    """Compression of an entire folder (e.g., circuit) using the given compression file format.
    The following compression formats are available: gzip (.gz; default), bzip2 (.bz2), LZMA (.xz)
    """

    FILE_FORMATS: ClassVar[tuple[str, ...]] = ("gz", "bz2", "xz")  # Supported compression formats

    def run(self) -> None:
        try:
            # Initial checks
            assert os.path.isdir(self.initialize.folder_path.path), (
                f"Folder path '{self.initialize.folder_path}' is not a valid directory!"
            )
            assert self.initialize.folder_path.path[-1] != os.path.sep, (
                f"Please remove trailing separator '{os.path.sep}' from path!"
            )
            assert self.initialize.file_format in self.FILE_FORMATS, (
                f"File format '{self.initialize.file_format}' not supported! Supported formats: {self.FILE_FORMATS}"
            )

            output_file = os.path.join(
                self.coordinate_output_root,
                f"{self.initialize.file_name}.{self.initialize.file_format}",
            )
            assert not os.path.exists(output_file), f"Output file '{output_file}' already exists!"

            # Compress using specified file format
            L.info(
                f"Info: Running {self.initialize.file_format} compression on '{self.initialize.folder_path}'...",
            )
            t0 = time.time()
            with tarfile.open(output_file, f"w:{self.initialize.file_format}") as tar:
                tar.add(
                    self.initialize.folder_path.path,
                    arcname=os.path.basename(self.initialize.folder_path.path),
                )

            # Once done, check elapsed time and resulting file size for reporting
            dt = time.time() - t0
            t_str = time.strftime("%Hh:%Mmin:%Ss", time.gmtime(dt))
            file_size = os.stat(output_file).st_size / (1024 * 1024)  # (MB)
            if file_size < 1024:
                file_unit = "MB"
            else:
                file_size = file_size / 1024
                file_unit = "GB"
            L.info(f"DONE (Duration {t_str}; File size {file_size:.1f}{file_unit})", flush=True)

        except Exception as e:
            traceback.print_exception(e)



from pathlib import Path
from typing import ClassVar

import morphio
import neurom
from entitysdk.models import ReconstructionMorphology
from entitysdk.models.entity import Entity
from pydantic import PrivateAttr

from obi_one.database.db_manager import db
from obi_one.database.entity_from_id import EntityFromID, LoadAssetMethod

import io

class ReconstructionMorphologyFromID(EntityFromID):
    entitysdk_class: ClassVar[type[Entity]] = ReconstructionMorphology
    _entity: ReconstructionMorphology | None = PrivateAttr(default=None)
    _swc_file_path: Path | None = PrivateAttr(default=None)
    _neurom_morphology: neurom.core.Morphology | None = PrivateAttr(default=None)
    _morphio_morphology: morphio.Morphology | None = PrivateAttr(default=None)
    _swc_file_content: str | None = PrivateAttr(default=None) 

    # @property
    def swc_file_content(self, db_client) -> None:
        """Function for downloading SWC files of a morphology into memory."""
        
        
        if self._swc_file_content is None:
            for asset in self.entity(db_client=db_client).assets:
                if asset.content_type == "application/swc":

                    load_asset_method = LoadAssetMethod.MEMORY
                    if load_asset_method == LoadAssetMethod.MEMORY:
                        print("Downloading SWC file for morphology...")

                        # Download the content into memory
                        content = db_client.download_content(
                            entity_id=self.entity(db_client=db_client).id,
                            entity_type=self.entitysdk_type,
                            asset_id=asset.id,
                        ).decode(encoding="utf-8")

                        type(content)

                        self._swc_file_content = content
                        break

                    #     # Use StringIO to create a file-like object in memory from the string content
                    #     neurom_morphology = neurom.load_morphology(io.StringIO(content), reader="asc")

                    # else:
                    #     file_output_path = Path(db.entity_file_store_path) / asset.full_path
                    #     file_output_path.parent.mkdir(parents=True, exist_ok=True)

                    #     # db.client.download_file(
                    #     client.download_file(
                    #         entity_id=self.entity(client).id,
                    #         entity_type=self.entitysdk_type,
                    #         asset_id=asset.id,
                    #         output_path=file_output_path,
                    #         # token=db.token,
                    #     )

                    #     self._swc_file_path = file_output_path
                    # break

            if self._swc_file_content is None:
                msg = "No valid application/asc asset found for morphology."
                raise ValueError(msg)

        return self._swc_file_content

    # @property
    def neurom_morphology(self, db_client) -> neurom.core.Morphology:
        """Getter for the neurom_morphology property.

        Downloads the application/asc asset if not already downloaded
        and loads it using neurom.load_morphology.
        """
        
        if self._neurom_morphology is None:
            self._neurom_morphology = neurom.load_morphology(io.StringIO(self.swc_file_content(db_client)), reader="swc")
        return self._neurom_morphology

    # # @property
    # def morphio_morphology(self, db_client) -> morphio.Morphology:
    #     """Getter for the morphio_morphology property.

    #     Downloads the application/asc asset if not already downloaded
    #     and initializes it as morphio.Morphology([...]).
    #     """
    #     if self._morphio_morphology is None:
    #         self._morphio_morphology = morphio.Morphology(io.StringIO(self.swc_file_content(db_client)), reader="asc")
    #     return self._morphio_morphology
